import pytest
import torch

from agilerl.modules.resnet import EvolvableResNet
from tests.helper_functions import assert_state_dicts_equal


######### Test instantiation #########
@pytest.mark.parametrize(
    "input_shape, channel_size, kernel_size, stride_size, num_outputs, num_blocks",
    [
        ([3, 32, 32], 64, 3, 1, 10, 2),
        ([3, 64, 64], 128, 3, 2, 10, 3),
    ],
)
def test_instantiation_without_errors(
    input_shape,
    channel_size,
    kernel_size,
    stride_size,
    num_outputs,
    num_blocks,
    device,
):
    evolvable_resnet = EvolvableResNet(
        input_shape=input_shape,
        channel_size=channel_size,
        kernel_size=kernel_size,
        stride_size=stride_size,
        num_outputs=num_outputs,
        num_blocks=num_blocks,
        device=device,
    )
    assert isinstance(evolvable_resnet, EvolvableResNet)


@pytest.mark.parametrize(
    "input_shape, channel_size, kernel_size, stride_size, num_outputs, num_blocks",
    [
        ([3, 32, 32], 64, 3, 1, 10, "two"),  # Incorrect type for num_blocks
        ([3, 32, 32], 64, 3, 1, 0, 2),  # num_outputs cannot be zero
    ],
)
def test_incorrect_instantiation(
    input_shape,
    channel_size,
    kernel_size,
    stride_size,
    num_outputs,
    num_blocks,
    device,
):
    with pytest.raises(AssertionError):
        EvolvableResNet(
            input_shape=input_shape,
            channel_size=channel_size,
            kernel_size=kernel_size,
            stride_size=stride_size,
            num_outputs=num_outputs,
            num_blocks=num_blocks,
            device=device,
        )


######### Test forward #########
@pytest.mark.parametrize(
    "input_shape, channel_size, kernel_size, stride_size, num_outputs, num_blocks, output_shape",
    [
        ([3, 32, 32], 64, 3, 1, 10, 2, (1, 10)),
        ([3, 64, 64], 128, 3, 2, 10, 3, (1, 10)),
    ],
)
def test_forward(
    input_shape,
    channel_size,
    kernel_size,
    stride_size,
    num_outputs,
    num_blocks,
    output_shape,
    device,
):
    evolvable_resnet = EvolvableResNet(
        input_shape=input_shape,
        channel_size=channel_size,
        kernel_size=kernel_size,
        stride_size=stride_size,
        num_outputs=num_outputs,
        num_blocks=num_blocks,
        device=device,
    )
    input_tensor = torch.randn(1, *input_shape).to(device)
    output = evolvable_resnet.forward(input_tensor)
    assert output.shape == output_shape


######### Test add_block #########
@pytest.mark.parametrize(
    "input_shape, channel_size, kernel_size, stride_size, num_outputs, num_blocks",
    [
        ([3, 32, 32], 64, 3, 1, 10, 2),
        ([3, 64, 64], 128, 3, 2, 10, 3),
    ],
)
def test_add_block(
    input_shape,
    channel_size,
    kernel_size,
    stride_size,
    num_outputs,
    num_blocks,
    device,
):
    evolvable_resnet = EvolvableResNet(
        input_shape=input_shape,
        channel_size=channel_size,
        kernel_size=kernel_size,
        stride_size=stride_size,
        num_outputs=num_outputs,
        num_blocks=num_blocks,
        device=device,
    )
    initial_blocks = evolvable_resnet.num_blocks
    evolvable_resnet.add_block()
    assert evolvable_resnet.num_blocks == initial_blocks + 1


######### Test remove_block #########
@pytest.mark.parametrize(
    "input_shape, channel_size, kernel_size, stride_size, num_outputs, num_blocks",
    [
        ([3, 32, 32], 64, 3, 1, 10, 2),
        ([3, 64, 64], 128, 3, 2, 10, 3),
    ],
)
def test_remove_block(
    input_shape,
    channel_size,
    kernel_size,
    stride_size,
    num_outputs,
    num_blocks,
    device,
):
    evolvable_resnet = EvolvableResNet(
        input_shape=input_shape,
        channel_size=channel_size,
        kernel_size=kernel_size,
        stride_size=stride_size,
        num_outputs=num_outputs,
        num_blocks=num_blocks,
        device=device,
    )
    initial_blocks = evolvable_resnet.num_blocks
    evolvable_resnet.remove_block()
    assert evolvable_resnet.num_blocks == initial_blocks - 1


######### Test add_channel #########
@pytest.mark.parametrize(
    "input_shape, channel_size, kernel_size, stride_size, num_outputs, num_blocks",
    [
        ([3, 32, 32], 64, 3, 1, 10, 2),
        ([3, 64, 64], 128, 3, 2, 10, 3),
    ],
)
def test_add_channel(
    input_shape,
    channel_size,
    kernel_size,
    stride_size,
    num_outputs,
    num_blocks,
    device,
):
    evolvable_resnet = EvolvableResNet(
        input_shape=input_shape,
        channel_size=channel_size,
        kernel_size=kernel_size,
        stride_size=stride_size,
        num_outputs=num_outputs,
        num_blocks=num_blocks,
        device=device,
    )
    initial_channel_size = evolvable_resnet.channel_size
    result = evolvable_resnet.add_channel()
    numb_new_channels = result["numb_new_channels"]
    assert evolvable_resnet.channel_size == initial_channel_size + numb_new_channels


######### Test remove_channel #########
@pytest.mark.parametrize(
    "input_shape, channel_size, kernel_size, stride_size, num_outputs, num_blocks",
    [
        ([3, 32, 32], 128, 3, 1, 10, 2),
        ([3, 64, 64], 128, 3, 2, 10, 3),
    ],
)
def test_remove_channel(
    input_shape,
    channel_size,
    kernel_size,
    stride_size,
    num_outputs,
    num_blocks,
    device,
):
    evolvable_resnet = EvolvableResNet(
        input_shape=input_shape,
        channel_size=channel_size,
        kernel_size=kernel_size,
        stride_size=stride_size,
        num_outputs=num_outputs,
        num_blocks=num_blocks,
        device=device,
    )
    initial_channel_size = evolvable_resnet.channel_size
    result = evolvable_resnet.remove_channel()
    numb_new_channels = result["numb_new_channels"]
    assert evolvable_resnet.channel_size == initial_channel_size - numb_new_channels


######### Test clone #########
@pytest.mark.parametrize(
    "input_shape, channel_size, kernel_size, stride_size, num_outputs, num_blocks",
    [
        ([3, 32, 32], 64, 3, 1, 10, 2),
        ([3, 64, 64], 128, 3, 2, 10, 3),
    ],
)
def test_clone_instance(
    input_shape,
    channel_size,
    kernel_size,
    stride_size,
    num_outputs,
    num_blocks,
    device,
):
    evolvable_resnet = EvolvableResNet(
        input_shape=input_shape,
        channel_size=channel_size,
        kernel_size=kernel_size,
        stride_size=stride_size,
        num_outputs=num_outputs,
        num_blocks=num_blocks,
        device=device,
    )
    original_net_dict = dict(evolvable_resnet.model.named_parameters())
    clone = evolvable_resnet.clone()
    clone_net = clone.model
    assert isinstance(clone, EvolvableResNet)
    assert_state_dicts_equal(clone.state_dict(), evolvable_resnet.state_dict())
    for key, param in clone_net.named_parameters():
        torch.testing.assert_close(param, original_net_dict[key])



import copy
import gc

import numpy as np
import pytest
import torch
from accelerate import Accelerator
from accelerate.state import AcceleratorState
from accelerate.utils import DeepSpeedPlugin
from gymnasium import spaces
from peft import LoraConfig

from agilerl.algorithms.core import EvolvableAlgorithm
from agilerl.algorithms.core.registry import HyperparameterConfig, RLParameter
from agilerl.hpo.mutation import MutationError, Mutations
from agilerl.modules import EvolvableBERT, ModuleDict
from agilerl.utils.utils import create_population
from agilerl.wrappers.agent import AsyncAgentsWrapper, RSNorm
from tests.helper_functions import assert_state_dicts_equal
from tests.test_algorithms.test_llms.test_grpo import create_module

# Shared HP dict that can be used by any algorithm
SHARED_INIT_HP = {
    "POPULATION_SIZE": 2,
    "DOUBLE": True,
    "BATCH_SIZE": 32,
    "CUDAGRAPHS": False,
    "LR": 1e-3,
    "LR_ACTOR": 1e-4,
    "LR_CRITIC": 1e-3,
    "GAMMA": 0.99,
    "LEARN_STEP": 1,
    "TAU": 1e-3,
    "BETA": 0.4,
    "PRIOR_EPS": 0.000001,
    "NUM_ATOMS": 51,
    "V_MIN": 0,
    "V_MAX": 200,
    "N_STEP": 3,
    "POLICY_FREQ": 10,
    "GAE_LAMBDA": 0.95,
    "ACTION_STD_INIT": 0.6,
    "CLIP_COEF": 0.2,
    "ENT_COEF": 0.01,
    "VF_COEF": 0.5,
    "MAX_GRAD_NORM": 0.5,
    "TARGET_KL": None,
    "UPDATE_EPOCHS": 4,
    "AGENT_IDS": ["agent_0", "agent_1", "other_agent_0"],
    "LAMBDA": 1.0,
    "REG": 0.000625,
    "CHANNELS_LAST": False,
    "O_U_NOISE": True,
    "EXPL_NOISE": 0.1,
    "MEAN_NOISE": 0.0,
    "THETA": 0.15,
    "DT": 0.01,
}

SHARED_INIT_HP_MA = SHARED_INIT_HP.copy()


def create_bert_network(device):
    return EvolvableBERT([12], [12], device=device)


def create_bert_networks_multi_agent(device):
    return ModuleDict(
        {
            "agent_0": create_bert_network(device),
            "agent_1": create_bert_network(device),
            "other_agent_0": create_bert_network(device),
        }
    )


@pytest.fixture(scope="function")
def bert_network(device):
    return create_bert_network(device)


@pytest.fixture(scope="function")
def bert_networks_multi_agent(device):
    return create_bert_networks_multi_agent(device)


@pytest.fixture(scope="function")
def bert_matd3_critic_networks(device):
    return [
        create_bert_networks_multi_agent(device),
        create_bert_networks_multi_agent(device),
    ]


@pytest.fixture(scope="function")
def init_pop(
    algo,
    observation_space,
    action_space,
    net_config,
    INIT_HP,
    population_size,
    device,
    accelerator,
    hp_config,
    torch_compiler,
    request,
    actor_network=None,
    critic_network=None,
):
    observation_space = request.getfixturevalue(observation_space)
    action_space = request.getfixturevalue(action_space)

    if hp_config is not None:
        hp_config = request.getfixturevalue(hp_config)

    if actor_network is not None:
        actor_network = request.getfixturevalue(actor_network)
    if critic_network is not None:
        critic_network = request.getfixturevalue(critic_network)

    pop = create_population(
        algo=algo,
        observation_space=observation_space,
        action_space=action_space,
        hp_config=hp_config,
        net_config=request.getfixturevalue(net_config),
        INIT_HP=INIT_HP,
        population_size=population_size,
        device=device,
        accelerator=accelerator,
        actor_network=actor_network,
        critic_network=critic_network,
        torch_compiler=torch_compiler,
    )
    yield pop
    for agent in pop:
        del agent
    del pop
    gc.collect()


# The constructor initializes all the attributes of the Mutations class correctly.
def test_constructor_initializes_attributes():
    no_mutation = 0.1
    architecture = 0.2
    new_layer_prob = 0.3
    parameters = 0.4
    activation = 0.5
    rl_hp = 0.6
    mutation_sd = 0.7
    activation_selection = ["ReLU", "Sigmoid"]
    mutate_elite = True
    rand_seed = 12345
    device = "cpu"
    accelerator = None

    mutations = Mutations(
        no_mutation,
        architecture,
        new_layer_prob,
        parameters,
        activation,
        rl_hp,
        mutation_sd,
        activation_selection,
        mutate_elite,
        rand_seed,
        device,
        accelerator,
    )

    assert mutations.rng is not None
    assert mutations.no_mut == no_mutation
    assert mutations.architecture_mut == architecture
    assert mutations.new_layer_prob == new_layer_prob
    assert mutations.parameters_mut == parameters
    assert mutations.activation_mut == activation
    assert mutations.rl_hp_mut == rl_hp
    assert mutations.mutation_sd == mutation_sd
    assert mutations.activation_selection == activation_selection
    assert mutations.mutate_elite == mutate_elite
    assert mutations.device == device
    assert mutations.accelerator == accelerator

    del mutations


# Checks no mutations if all probabilities set to zero
@pytest.mark.parametrize("algo", ["DQN"])
@pytest.mark.parametrize(
    "observation_space, net_config",
    [("vector_space", "encoder_mlp_config")],
)
@pytest.mark.parametrize("action_space", ["discrete_space"])
@pytest.mark.parametrize("accelerator", [None])
@pytest.mark.parametrize("torch_compiler", [None])
@pytest.mark.parametrize("INIT_HP", [SHARED_INIT_HP])
@pytest.mark.parametrize("population_size", [1])
@pytest.mark.parametrize("hp_config", ["default_hp_config"])
def test_mutation_no_options(init_pop, device):
    pre_training_mut = True
    population = init_pop
    mutations = Mutations(0, 0, 0, 0, 0, 0, 0.1, device=device)

    new_population = [agent.clone() for agent in population]
    mutated_population = mutations.mutation(new_population, pre_training_mut)

    assert len(mutated_population) == len(population)
    for old, individual in zip(population, mutated_population):
        assert_state_dicts_equal(old.actor.state_dict(), individual.actor.state_dict())

    del mutations, mutated_population, new_population


#### Single-agent algorithm mutations ####
# The mutation method applies random mutations to the population and returns the mutated population.
@pytest.mark.parametrize(
    "algo, hp_config, action_space",
    [
        ("DQN", "default_hp_config", "discrete_space"),
        ("Rainbow DQN", "default_hp_config", "discrete_space"),
        ("DDPG", "ac_hp_config", "vector_space"),
        ("TD3", "ac_hp_config", "vector_space"),
        ("PPO", "default_hp_config", "discrete_space"),
        ("CQN", "default_hp_config", "discrete_space"),
        ("NeuralUCB", "default_hp_config", "discrete_space"),
        ("NeuralTS", "default_hp_config", "discrete_space"),
    ],
)
@pytest.mark.parametrize("torch_compiler", [None])
@pytest.mark.parametrize("accelerator", [None, Accelerator(device_placement=False)])
@pytest.mark.parametrize("INIT_HP", [SHARED_INIT_HP])
@pytest.mark.parametrize(
    "observation_space, net_config", [("vector_space", "encoder_mlp_config")]
)
@pytest.mark.parametrize("population_size", [1])
def test_mutation_applies_random_mutations(algo, init_pop, device, accelerator):
    population = init_pop
    pre_training_mut = True

    population = init_pop

    mutations = Mutations(
        0,
        0.1,
        0.1,
        0.1,
        0.1,
        0.1,
        0.1,
        mutate_elite=False,
        device=device,
        accelerator=accelerator,
    )

    # Unwrap models if using accelerator
    if accelerator is not None:
        for agent in population:
            agent.unwrap_models()

    mutated_population = mutations.mutation(population, pre_training_mut)

    assert len(mutated_population) == len(population)
    assert mutated_population[0].mut == "None"  # Satisfies mutate_elite=False condition
    for individual in mutated_population:
        policy = getattr(individual, individual.registry.policy())
        assert individual.mut in [
            "None",
            "batch_size",
            "lr",
            "lr_actor",
            "lr_critic",
            "learn_step",
            "act",
            "param",
            policy.last_mutation_attr,
        ]

    del mutations, mutated_population


# The mutation method applies no mutations to the population and returns the mutated population.
@pytest.mark.parametrize(
    "algo, action_space",
    [
        ("DQN", "discrete_space"),
        ("Rainbow DQN", "discrete_space"),
        ("DDPG", "vector_space"),
        ("TD3", "vector_space"),
        ("PPO", "discrete_space"),
        ("CQN", "discrete_space"),
        ("NeuralUCB", "discrete_space"),
        ("NeuralTS", "discrete_space"),
    ],
)
@pytest.mark.parametrize(
    "observation_space, net_config", [("vector_space", "encoder_mlp_config")]
)
@pytest.mark.parametrize("torch_compiler", [None])
@pytest.mark.parametrize("accelerator", [None, Accelerator(device_placement=False)])
@pytest.mark.parametrize("INIT_HP", [SHARED_INIT_HP])
@pytest.mark.parametrize("hp_config", [None])
@pytest.mark.parametrize("population_size", [1])
def test_mutation_applies_no_mutations(init_pop, device, accelerator):
    pre_training_mut = False

    population = init_pop

    mutations = Mutations(
        1,
        0,
        0,
        0,
        0,
        0,
        0.1,
        device=device,
        accelerator=accelerator,
    )

    new_population = [agent.clone(wrap=False) for agent in population]
    mutated_population = mutations.mutation(new_population, pre_training_mut)

    assert len(mutated_population) == len(population)
    for old, individual in zip(population, mutated_population):
        assert individual.mut in ["None"]
        assert old.index == individual.index
        assert old.actor != individual.actor
        assert_state_dicts_equal(old.actor.state_dict(), individual.actor.state_dict())

    del mutations, mutated_population, new_population


# The mutation method applies no mutations to the population and returns the mutated population.
@pytest.mark.parametrize(
    "algo, action_space",
    [
        ("DQN", "discrete_space"),
        ("Rainbow DQN", "discrete_space"),
        ("DDPG", "vector_space"),
        ("TD3", "vector_space"),
        ("PPO", "discrete_space"),
        ("CQN", "discrete_space"),
        ("NeuralUCB", "discrete_space"),
        ("NeuralTS", "discrete_space"),
    ],
)
@pytest.mark.parametrize(
    "observation_space, net_config", [("vector_space", "encoder_mlp_config")]
)
@pytest.mark.parametrize("accelerator", [None, Accelerator(device_placement=False)])
@pytest.mark.parametrize("INIT_HP", [SHARED_INIT_HP])
@pytest.mark.parametrize("torch_compiler", [None])
@pytest.mark.parametrize("hp_config", [None])
@pytest.mark.parametrize("population_size", [1])
def test_mutation_applies_no_mutations_pre_training_mut(init_pop, device, accelerator):
    pre_training_mut = True
    population = init_pop

    # Set all mutation probabilities to 0
    mutations = Mutations(
        1,
        0,
        0,
        0,
        0,
        1,
        0.1,
        device=device,
        accelerator=accelerator,
    )

    new_population = [agent.clone(wrap=False) for agent in population]
    mutated_population = mutations.mutation(new_population, pre_training_mut)

    assert len(mutated_population) == len(population)
    for old, individual in zip(population, mutated_population):
        assert individual.mut in [
            "None",
            "batch_size",
            "lr",
            "lr_actor",
            "lr_critic",
            "learn_step",
        ]
        assert old.index == individual.index
        assert old.actor != individual.actor
        assert_state_dicts_equal(old.actor.state_dict(), individual.actor.state_dict())

    del mutations, mutated_population, new_population


# The mutation method applies RL hyperparameter mutations to the population and returns the mutated population.
@pytest.mark.parametrize(
    "algo, hp_config, action_space",
    [
        ("DQN", "default_hp_config", "discrete_space"),
        ("Rainbow DQN", "default_hp_config", "discrete_space"),
        ("DDPG", "ac_hp_config", "vector_space"),
        ("TD3", "ac_hp_config", "vector_space"),
        ("PPO", "default_hp_config", "discrete_space"),
        ("CQN", "default_hp_config", "discrete_space"),
        ("NeuralUCB", "default_hp_config", "discrete_space"),
        ("NeuralTS", "default_hp_config", "discrete_space"),
    ],
)
@pytest.mark.parametrize(
    "observation_space, net_config", [("vector_space", "encoder_mlp_config")]
)
@pytest.mark.parametrize("torch_compiler", [None])
@pytest.mark.parametrize("accelerator", [None, Accelerator(device_placement=False)])
@pytest.mark.parametrize("INIT_HP", [SHARED_INIT_HP])
@pytest.mark.parametrize("population_size", [1])
def test_mutation_applies_rl_hp_mutations(
    init_pop, device, accelerator, hp_config, request
):
    pre_training_mut = False
    population = init_pop
    mutations = Mutations(
        0,
        0,
        0,
        0,
        0,
        1,
        0.1,
        device=device,
        accelerator=accelerator,
    )
    hp_config = request.getfixturevalue(hp_config)

    new_population = [agent.clone(wrap=False) for agent in population]
    mutated_population = mutations.mutation(new_population, pre_training_mut)

    assert len(mutated_population) == len(population)
    for old, individual in zip(population, mutated_population):
        available_mutations = hp_config.names()
        assert individual.mut in available_mutations

        new_value = getattr(individual, individual.mut)
        min_value = hp_config[individual.mut].min
        max_value = hp_config[individual.mut].max
        assert min_value <= new_value <= max_value
        assert old.index == individual.index

    del mutations, mutated_population, new_population


# The mutation method applies activation mutations to the population and returns the mutated population.
@pytest.mark.parametrize(
    "algo, action_space",
    [
        ("DQN", "discrete_space"),
        ("Rainbow DQN", "discrete_space"),
        ("DDPG", "vector_space"),
        ("TD3", "vector_space"),
        ("PPO", "discrete_space"),
        ("CQN", "discrete_space"),
        ("NeuralUCB", "discrete_space"),
        ("NeuralTS", "discrete_space"),
    ],
)
@pytest.mark.parametrize(
    "observation_space, net_config",
    [
        ("vector_space", "encoder_mlp_config"),
        ("image_space", "encoder_cnn_config"),
        ("dict_space", "encoder_multi_input_config"),
    ],
)
@pytest.mark.parametrize("torch_compiler", [None])
@pytest.mark.parametrize("accelerator", [None, Accelerator(device_placement=False)])
@pytest.mark.parametrize("INIT_HP", [SHARED_INIT_HP])
@pytest.mark.parametrize("hp_config", [None])
@pytest.mark.parametrize("population_size", [1])
def test_mutation_applies_activation_mutations(
    init_pop, observation_space, device, accelerator
):
    pre_training_mut = False
    population = init_pop

    if isinstance(observation_space, spaces.Box) and len(observation_space.shape) == 3:
        activation_selection = ["ReLU", "ELU", "GELU"]
    else:
        activation_selection = ["Tanh", "ReLU", "ELU", "GELU"]

    mutations = Mutations(
        0,
        0,
        0,
        0,
        1,
        0,
        0.1,
        activation_selection=activation_selection,
        device=device,
        accelerator=accelerator,
    )

    new_population = [agent.clone(wrap=False) for agent in population]
    mutated_population = mutations.mutation(new_population, pre_training_mut)

    assert len(mutated_population) == len(population)
    for old, individual in zip(population, mutated_population):
        assert individual.mut in ["None", "act"]
        if individual.mut == "act":
            assert old.actor.activation != individual.actor.activation
            assert individual.actor.activation in activation_selection
        assert old.index == individual.index

    del mutations, mutated_population, new_population


# The mutation method applies activation mutations to the population and returns the mutated population.
@pytest.mark.parametrize(
    "observation_space, net_config",
    [
        ("vector_space", "encoder_mlp_config"),
        ("image_space", "encoder_cnn_config"),
        ("dict_space", "encoder_multi_input_config"),
        ("discrete_space", "encoder_mlp_config"),
    ],
)
@pytest.mark.parametrize("algo, action_space", [("DDPG", "vector_space")])
@pytest.mark.parametrize("accelerator", [None, Accelerator(device_placement=False)])
@pytest.mark.parametrize("INIT_HP", [SHARED_INIT_HP])
@pytest.mark.parametrize("torch_compiler", [None])
@pytest.mark.parametrize("hp_config", [None])
@pytest.mark.parametrize("population_size", [1])
def test_mutation_applies_activation_mutations_no_skip(init_pop, device, accelerator):
    pre_training_mut = False
    population = init_pop
    mutations = Mutations(
        0,
        0,
        0,
        0,
        1,
        0,
        0.1,
        device=device,
        accelerator=accelerator,
    )

    for individual in population:
        individual.algo = None
        individual.lr = 1e-3
    new_population = [agent.clone(wrap=False) for agent in population]
    mutated_population = mutations.mutation(new_population, pre_training_mut)

    assert len(mutated_population) == len(population)
    for old, individual in zip(population, mutated_population):
        assert individual.mut in ["None", "act"]
        if individual.mut == "act":
            assert old.actor.activation != individual.actor.activation
            assert individual.actor.activation in ["ReLU", "ELU", "GELU"]
        assert old.index == individual.index

    del mutations, mutated_population, new_population


# The mutation method applies parameter mutations to the population and returns the mutated population.
@pytest.mark.parametrize(
    "algo, action_space, wrapper_cls",
    [
        ("DQN", "discrete_space", None),
        ("Rainbow DQN", "discrete_space", None),
        ("DDPG", "vector_space", None),
        ("DDPG", "vector_space", RSNorm),
        ("TD3", "vector_space", None),
        ("PPO", "discrete_space", None),
        ("CQN", "discrete_space", None),
        ("NeuralUCB", "discrete_space", None),
        ("NeuralTS", "discrete_space", None),
    ],
)
@pytest.mark.parametrize(
    "observation_space, net_config", [("vector_space", "encoder_mlp_config")]
)
@pytest.mark.parametrize("torch_compiler", [None])
@pytest.mark.parametrize("accelerator", [None, Accelerator(device_placement=False)])
@pytest.mark.parametrize("INIT_HP", [SHARED_INIT_HP])
@pytest.mark.parametrize("hp_config", [None])
@pytest.mark.parametrize("population_size", [1])
def test_mutation_applies_parameter_mutations(
    algo, device, accelerator, init_pop, wrapper_cls
):
    pre_training_mut = False

    population = init_pop

    if wrapper_cls is not None:
        population = [wrapper_cls(agent) for agent in population]

    mutations = Mutations(
        0,
        0,
        0,
        1,
        0,
        0,
        0.5,
        device=device,
        accelerator=accelerator,
    )

    new_population = [agent.clone(wrap=False) for agent in population]
    mutated_population = mutations.mutation(new_population, pre_training_mut)

    assert len(mutated_population) == len(population)
    for old, individual in zip(population, mutated_population):
        assert individual.mut == "param"
        # Due to randomness, sometimes parameters are not different
        # assert str(old.actor.state_dict()) != str(individual.actor.state_dict())
        assert old.index == individual.index

        # Compare state dictionaries of the actor (or network)
        policy_name = old.registry.policy()
        old_policy = getattr(old, policy_name)
        new_policy = getattr(individual, policy_name)
        old_sd = old_policy.state_dict()
        new_sd = new_policy.state_dict()
        mutation_found = False
        for key in old_sd.keys():
            if "norm" in key:  # Skip normalization layers
                continue
            diff_norm = (old_sd[key] - new_sd[key]).norm().item()
            if diff_norm > 1e-6:
                mutation_found = True
                break

        assert mutation_found, f"Mutation not applied for agent index {old.index}"

    del mutations, mutated_population, new_population


# The mutation method applies architecture mutations to the population and returns the mutated population.
@pytest.mark.parametrize(
    "algo, action_space, wrapper_cls",
    [
        ("DQN", "discrete_space", None),
        ("Rainbow DQN", "discrete_space", None),
        ("DDPG", "vector_space", None),
        ("DDPG", "vector_space", RSNorm),
        ("TD3", "vector_space", None),
        ("PPO", "discrete_space", None),
        ("CQN", "discrete_space", None),
        ("NeuralUCB", "discrete_space", None),
        ("NeuralTS", "discrete_space", None),
    ],
)
@pytest.mark.parametrize(
    "observation_space, net_config",
    [
        ("vector_space", "encoder_mlp_config"),
        ("image_space", "encoder_cnn_config"),
        ("dict_space", "encoder_multi_input_config"),
    ],
)
@pytest.mark.parametrize("torch_compiler", [None])
@pytest.mark.parametrize("accelerator", [None, Accelerator(device_placement=False)])
@pytest.mark.parametrize("INIT_HP", [SHARED_INIT_HP])
@pytest.mark.parametrize("hp_config", [None])
@pytest.mark.parametrize("population_size", [1])
def test_mutation_applies_architecture_mutations(
    init_pop, device, accelerator, wrapper_cls
):
    population: list[EvolvableAlgorithm] = init_pop
    if wrapper_cls is not None:
        population = [wrapper_cls(agent) for agent in population]

    mutations = Mutations(
        0,
        1,
        0.5,
        0,
        0,
        0,
        0.5,
        device=device,
        accelerator=accelerator,
    )

    mut_methods = population[0].actor.mutation_methods

    # Change EvolvableModule random number generator to test mutation methods
    class EvoDummyRNG:
        rng = np.random.default_rng(seed=42)

        def choice(self, a, size=None, replace=True, p=None):
            return 1

        def integers(self, low=0, high=None):
            return self.rng.integers(low, high)

    for individual in population:
        for name, network in individual.evolvable_attributes(
            networks_only=True
        ).items():
            network.rng = EvoDummyRNG()

    applied_mutations = set()
    for mut_method in mut_methods:

        class DummyRNG:
            def choice(self, a, size=None, replace=True, p=None):
                return [mut_method]

        mutations.rng = DummyRNG()

        new_population = [agent.clone(wrap=False) for agent in population]

        # Apply architecture mutations to the population
        if isinstance(population[0], RSNorm):
            new_population = [agent.agent for agent in new_population]

        mutated_population = [
            mutations.architecture_mutate(agent) for agent in new_population
        ]
        for individual in mutated_population:
            individual.mutation_hook()

        assert len(mutated_population) == len(population)
        for old, individual in zip(population, mutated_population):
            policy_name = old.registry.policy()
            policy = getattr(individual, policy_name)
            # old_policy = getattr(old, policy_name)
            assert individual.mut == (policy.last_mutation_attr or "None")

            if policy.last_mutation_attr is not None:
                applied_mutations.add(policy.last_mutation_attr)
                # assert str(old_policy.state_dict()) != str(policy.state_dict())
                for group in old.registry.groups:
                    if group.eval_network != policy_name:
                        eval_module = getattr(individual, group.eval_network)
                        # old_eval_module = getattr(old, group.eval_network)
                        assert eval_module.last_mutation_attr is not None
                        assert (
                            eval_module.last_mutation_attr == policy.last_mutation_attr
                        )
                        # assert str(old_eval_module.state_dict()) != str(eval_module.state_dict())

            assert old.index == individual.index

        # assert_equal_state_dict(population, mutated_population)

    assert all(mut in mut_methods for mut in applied_mutations), set(mut_methods) - set(
        applied_mutations
    )

    del mutations, mutated_population, new_population


# The mutation method applies BERT architecture mutations to the population and returns the mutated population.
@pytest.mark.skip(reason="Skipping BERT architecture mutations test.")
@pytest.mark.parametrize(
    "algo, actor_network, critic_network", [("DDPG", "bert_network", "bert_network")]
)
@pytest.mark.parametrize(
    "observation_space, net_config", [("vector_space", "encoder_mlp_config")]
)
@pytest.mark.parametrize("action_space", ["vector_space"])
@pytest.mark.parametrize("INIT_HP", [SHARED_INIT_HP])
@pytest.mark.parametrize("hp_config", [None])
@pytest.mark.parametrize("population_size", [1])
@pytest.mark.parametrize("torch_compiler", [None])
@pytest.mark.parametrize("accelerator", [None, Accelerator(device_placement=True)])
@pytest.mark.parametrize(
    "mut_method",
    [
        [
            "add_encoder_layer",
            "remove_encoder_layer",
            "add_decoder_layer",
            "remove_decoder_layer",
        ],
        ["add_node", "remove_node"],
    ],
)
def test_mutation_applies_bert_architecture_mutations_single_agent(
    algo,
    observation_space,
    action_space,
    device,
    accelerator,
    mut_method,
    actor_network,
    critic_network,
    init_pop,
    request,
):
    observation_space = request.getfixturevalue(observation_space)
    action_space = request.getfixturevalue(action_space)

    # Pass the network parameters to init_pop through the test
    actual_actor_network = (
        request.getfixturevalue(actor_network) if actor_network else None
    )
    actual_critic_network = (
        request.getfixturevalue(critic_network) if critic_network else None
    )

    # Create a custom population with the BERT networks
    from agilerl.utils.utils import create_population

    population = create_population(
        algo=algo,
        observation_space=observation_space,
        action_space=action_space,
        hp_config=None,
        net_config=request.getfixturevalue("encoder_mlp_config"),
        INIT_HP=SHARED_INIT_HP,
        population_size=1,
        device=device,
        accelerator=accelerator,
        actor_network=actual_actor_network,
        critic_network=actual_critic_network,
    )

    mutations = Mutations(
        0,
        1,
        0.5,
        0,
        0,
        0,
        0.5,
        device=device,
        accelerator=accelerator,
    )

    class DummyRNG:
        def choice(self, a, size=None, replace=True, p=None):
            return [np.random.choice(mut_method)]

    mutations.rng = DummyRNG()

    new_population = [agent.clone(wrap=False) for agent in population]
    mutated_population = [
        mutations.architecture_mutate(agent) for agent in new_population
    ]

    assert len(mutated_population) == len(population)
    for old, individual in zip(population, mutated_population):
        policy = getattr(individual, individual.registry.policy())
        assert individual.mut == policy.last_mutation_attr
        # Due to randomness and constraints on size, sometimes architectures are not different
        # assert str(old.actor.state_dict()) != str(individual.actor.state_dict())
        assert old.index == individual.index

    # assert_equal_state_dict(population, mutated_population)

    del mutations, mutated_population, new_population


#### Multi-agent algorithm mutations ####
# The mutation method applies random mutations to the population and returns the mutated population.
@pytest.mark.parametrize("algo", ["MADDPG", "MATD3", "IPPO"])
@pytest.mark.parametrize(
    "observation_space, net_config", [("ma_vector_space", "encoder_mlp_config")]
)
@pytest.mark.parametrize("action_space", ["ma_discrete_space"])
@pytest.mark.parametrize("hp_config", [None])
@pytest.mark.parametrize("population_size", [1])
@pytest.mark.parametrize("torch_compiler", [None])
@pytest.mark.parametrize("accelerator", [None, Accelerator(device_placement=False)])
@pytest.mark.parametrize("INIT_HP", [SHARED_INIT_HP_MA])
def test_mutation_applies_random_mutations_multi_agent(init_pop, device, accelerator):
    pre_training_mut = False
    population = init_pop

    # Random mutations
    mutations = Mutations(
        0,
        0.1,
        0.1,
        0.1,
        0.1,
        0.1,
        0.1,
        device=device,
        accelerator=accelerator,
    )

    if accelerator is not None:
        for agent in population:
            agent.unwrap_models()

    mutated_population = mutations.mutation(population, pre_training_mut)

    assert len(mutated_population) == len(population)
    for individual in mutated_population:
        policy = getattr(individual, individual.registry.policy())
        if policy.last_mutation_attr is not None:
            sampled_mutation = ".".join(policy.last_mutation_attr.split(".")[1:])
        else:
            sampled_mutation = "None"

        assert individual.mut in [
            "None",
            "batch_size",
            "lr",
            "lr_actor",
            "lr_critic",
            "learn_step",
            "act",
            "param",
            sampled_mutation,
        ]

    del mutations, mutated_population


# The mutation method applies no mutations to the population and returns the mutated population.
@pytest.mark.parametrize("algo", ["MADDPG", "MATD3", "IPPO"])
@pytest.mark.parametrize(
    "observation_space, net_config", [("ma_vector_space", "encoder_mlp_config")]
)
@pytest.mark.parametrize("action_space", ["ma_discrete_space"])
@pytest.mark.parametrize("INIT_HP", [SHARED_INIT_HP_MA])
@pytest.mark.parametrize("population_size", [1])
@pytest.mark.parametrize("hp_config", [None])
@pytest.mark.parametrize("torch_compiler", [None])
@pytest.mark.parametrize("accelerator", [None, Accelerator(device_placement=False)])
def test_mutation_applies_no_mutations_multi_agent(init_pop, device, accelerator):
    pre_training_mut = False
    population = init_pop

    mutations = Mutations(
        1,
        0,
        0,
        0,
        0,
        0,
        0.1,
        device=device,
        accelerator=accelerator,
    )

    if accelerator is not None:
        for agent in population:
            agent.unwrap_models()

    mutated_population = mutations.mutation(population, pre_training_mut)

    assert len(mutated_population) == len(population)
    for old, individual in zip(population, mutated_population):
        assert individual.mut in ["None"]
        assert old.index == individual.index
        assert old.actors == individual.actors

    del mutations
    del population
    del mutated_population


# The mutation method applies RL hyperparameter mutations to the population and returns the mutated population.
@pytest.mark.parametrize(
    "algo, hp_config",
    [
        ("MADDPG", "ac_hp_config"),
        ("MATD3", "ac_hp_config"),
        ("IPPO", "default_hp_config"),
    ],
)
@pytest.mark.parametrize(
    "observation_space, net_config", [("ma_vector_space", "encoder_mlp_config")]
)
@pytest.mark.parametrize("action_space", ["ma_discrete_space"])
@pytest.mark.parametrize("population_size", [1])
@pytest.mark.parametrize("torch_compiler", [None])
@pytest.mark.parametrize("accelerator", [None, Accelerator(device_placement=False)])
@pytest.mark.parametrize("INIT_HP", [SHARED_INIT_HP_MA])
def test_mutation_applies_rl_hp_mutations_multi_agent(
    init_pop, device, accelerator, hp_config, request
):
    pre_training_mut = False
    population = init_pop

    mutations = Mutations(
        0,
        0,
        0,
        0,
        0,
        1,
        0.1,
        device=device,
        accelerator=accelerator,
    )

    new_population = [agent.clone(wrap=False) for agent in population]
    mutated_population = mutations.mutation(new_population, pre_training_mut)

    hp_config = request.getfixturevalue(hp_config)

    assert len(mutated_population) == len(population)
    for old, individual in zip(population, mutated_population):
        available_mutations = hp_config.names()
        assert individual.mut in available_mutations

        new_value = getattr(individual, individual.mut)
        min_value = hp_config[individual.mut].min
        max_value = hp_config[individual.mut].max
        assert min_value <= new_value <= max_value
        assert old.index == individual.index

    del mutations, mutated_population, new_population


# The mutation method applies activation mutations to the population and returns the mutated population.
@pytest.mark.parametrize("algo", ["MADDPG", "MATD3", "IPPO"])
@pytest.mark.parametrize(
    "observation_space, net_config",
    [
        ("ma_vector_space", "encoder_mlp_config"),
        ("ma_image_space", "encoder_cnn_config"),
        ("ma_dict_space_small", "encoder_multi_input_config"),
    ],
)
@pytest.mark.parametrize("action_space", ["ma_discrete_space"])
@pytest.mark.parametrize("population_size", [1])
@pytest.mark.parametrize("hp_config", [None])
@pytest.mark.parametrize("torch_compiler", [None])
@pytest.mark.parametrize("accelerator", [None, Accelerator(device_placement=False)])
@pytest.mark.parametrize("INIT_HP", [SHARED_INIT_HP_MA])
def test_mutation_applies_activation_mutations_multi_agent(
    init_pop, device, accelerator
):
    pre_training_mut = False
    population = init_pop

    mutations = Mutations(
        0,
        0,
        0,
        0,
        1,
        0,
        0.1,
        device=device,
        accelerator=accelerator,
    )

    new_population = [agent.clone(wrap=False) for agent in population]
    mutated_population = mutations.mutation(new_population, pre_training_mut)

    assert len(mutated_population) == len(population)
    for old, individual in zip(population, mutated_population):
        assert individual.mut in ["None", "act"]
        if individual.mut == "act":
            for old_actor, actor in zip(old.actors, individual.actors):
                assert old_actor.activation != actor.activation
                assert individual.actors[0].activation in [
                    "ReLU",
                    "ELU",
                    "GELU",
                ]
        assert old.index == individual.index

    del mutations, mutated_population, new_population


# The mutation method applies activation mutations to the population and returns the mutated population.
@pytest.mark.parametrize("algo", ["MADDPG", "MATD3", "IPPO"])
@pytest.mark.parametrize(
    "observation_space, net_config", [("ma_vector_space", "encoder_mlp_config")]
)
@pytest.mark.parametrize("action_space", ["ma_discrete_space"])
@pytest.mark.parametrize("population_size", [1])
@pytest.mark.parametrize("hp_config", [None])
@pytest.mark.parametrize("torch_compiler", [None])
@pytest.mark.parametrize("accelerator", [None, Accelerator(device_placement=False)])
@pytest.mark.parametrize("INIT_HP", [SHARED_INIT_HP_MA])
def test_mutation_applies_activation_mutations_multi_agent_no_skip(
    init_pop, device, accelerator
):
    pre_training_mut = False
    population = init_pop

    mutations = Mutations(
        0,
        0,
        0,
        0,
        1,
        0,
        0.1,
        device=device,
        accelerator=accelerator,
    )

    for individual in population:
        individual.algo = None
    new_population = [agent.clone(wrap=False) for agent in population]
    mutated_population = mutations.mutation(new_population, pre_training_mut)

    assert len(mutated_population) == len(population)
    for old, individual in zip(population, mutated_population):
        assert individual.mut in ["None", "act"]
        if individual.mut == "act":
            for old_actor, actor in zip(
                old.actors.values(), individual.actors.values()
            ):
                assert old_actor.activation != actor.activation
                assert actor.activation in [
                    "ReLU",
                    "ELU",
                    "GELU",
                ]
        assert old.index == individual.index

    del mutations, mutated_population, new_population


# The mutation method applies parameter mutations to the population and returns the mutated population.
@pytest.mark.parametrize(
    "algo, wrapper_cls",
    [
        ("MADDPG", None),
        ("MATD3", None),
        ("IPPO", None),
        ("IPPO", AsyncAgentsWrapper),
    ],
)
@pytest.mark.parametrize(
    "observation_space, net_config", [("ma_vector_space", "encoder_mlp_config")]
)
@pytest.mark.parametrize("action_space", ["ma_discrete_space"])
@pytest.mark.parametrize("population_size", [1])
@pytest.mark.parametrize("hp_config", [None])
@pytest.mark.parametrize("torch_compiler", [None])
@pytest.mark.parametrize("accelerator", [None, Accelerator(device_placement=False)])
@pytest.mark.parametrize("INIT_HP", [SHARED_INIT_HP_MA])
def test_mutation_applies_parameter_mutations_multi_agent(
    init_pop, device, accelerator, wrapper_cls
):
    pre_training_mut = False
    population = init_pop

    if wrapper_cls is not None:
        population = [wrapper_cls(agent) for agent in population]

    mutations = Mutations(
        0,
        0,
        0,
        1,
        0,
        0,
        0.5,
        device=device,
        accelerator=accelerator,
    )

    new_population = [agent.clone(wrap=False) for agent in population]
    mutated_population = mutations.mutation(new_population, pre_training_mut)

    assert len(mutated_population) == len(population)
    for old, individual in zip(population, mutated_population):
        assert individual.mut == "param"
        # Due to randomness, sometimes parameters are not different
        # assert str(old.actors[0].state_dict()) != str(individual.actors[0].state_dict())
        assert old.index == individual.index

        # Compare state dictionaries of the actor (or network)
        policy_name = old.registry.policy()
        old_policy = getattr(old, policy_name)
        new_policy = getattr(individual, policy_name)
        old_sd = old_policy.state_dict()
        new_sd = new_policy.state_dict()
        mutation_found = False
        for key in old_sd.keys():
            if "norm" in key:  # Skip normalization layers
                continue
            diff_norm = (old_sd[key] - new_sd[key]).norm().item()
            if diff_norm > 1e-6:
                mutation_found = True
                break

        assert mutation_found, f"Mutation not applied for agent index {old.index}"

    del mutations, mutated_population, new_population


# The mutation method applies architecture mutations to the population and returns the mutated population.
@pytest.mark.parametrize(
    "algo, wrapper_cls",
    [
        ("MADDPG", None),
        ("MATD3", None),
        ("IPPO", None),
        ("IPPO", AsyncAgentsWrapper),
    ],
)
@pytest.mark.parametrize(
    "observation_space, net_config",
    [
        ("ma_vector_space", "encoder_mlp_config"),
        ("ma_image_space", "encoder_cnn_config"),
        # ("ma_dict_space_small", "encoder_multi_input_config"), NOTE: Takes too long to run
    ],
)
@pytest.mark.parametrize("action_space", ["ma_discrete_space"])
@pytest.mark.parametrize("population_size", [1])
@pytest.mark.parametrize("hp_config", [None])
@pytest.mark.parametrize("accelerator", [None, Accelerator(device_placement=False)])
@pytest.mark.parametrize("torch_compiler", [None, "default"])
@pytest.mark.parametrize("INIT_HP", [SHARED_INIT_HP_MA])
def test_mutation_applies_architecture_mutations_multi_agent(
    algo, init_pop, device, accelerator, wrapper_cls
):
    population: list[EvolvableAlgorithm] = init_pop
    mutations = Mutations(
        0,
        1,
        0.5,
        0,
        0,
        0,
        0.5,
        device=device,
        accelerator=accelerator,
    )

    # Change EvolvableModule random number generator to test mutation methods
    class EvoDummyRNG:
        rng = np.random.default_rng(seed=42)

        def choice(self, a, size=None, replace=True, p=None):
            return 1

        def integers(self, low=0, high=None):
            return self.rng.integers(low, high)

    if wrapper_cls is not None:
        population = [wrapper_cls(agent) for agent in population]

    for individual in population:
        for network in individual.evolvable_attributes(networks_only=True).values():
            network.rng = EvoDummyRNG()

    test_agent = "agent_0" if algo != "IPPO" else "agent"
    mut_methods = population[0].actors[test_agent].mutation_methods
    applied_mutations = set()
    for mut_method in mut_methods:

        class DummyRNG:
            def choice(self, a, size=None, replace=True, p=None):
                return [".".join([test_agent, mut_method])]

        mutations.rng = DummyRNG()

        new_population = [agent.clone(wrap=False) for agent in population]

        if isinstance(new_population[0], AsyncAgentsWrapper):
            new_population = [agent.agent for agent in new_population]

        mutated_population = [
            mutations.architecture_mutate(agent) for agent in new_population
        ]

        assert len(mutated_population) == len(population)
        for old, individual in zip(population, mutated_population):
            policy_name = individual.registry.policy()
            policy = getattr(individual, policy_name)
            # old_policy = getattr(old, policy_name)
            if policy.last_mutation_attr is not None:
                sampled_mutation = ".".join(policy.last_mutation_attr.split(".")[1:])
                applied_mutations.add(sampled_mutation)
            else:
                sampled_mutation = None

            assert individual.mut == sampled_mutation or "None"

            if sampled_mutation is not None:
                for group in old.registry.groups:
                    if group.eval_network != policy_name:
                        eval_module = getattr(individual, group.eval_network)
                        # old_eval_module = getattr(old, group.eval_network)
                        for _, module in eval_module.items():
                            bottom_eval_mut = module.last_mutation_attr.split(".")[-1]
                            bottom_policy_mut = policy.last_mutation_attr.split(".")[-1]
                            assert module.last_mutation_attr is not None
                            assert bottom_eval_mut == bottom_policy_mut

            assert old.index == individual.index

        del new_population, mutated_population

    del mutations

    assert all(mut in applied_mutations for mut in mut_methods), set(mut_methods) - set(
        applied_mutations
    )


# The mutation method applies BERT architecture mutations to the population and returns the mutated population.
@pytest.mark.skip(reason="Skipping BERT architecture mutations test.")
@pytest.mark.parametrize(
    "algo, actor_network, critic_network",
    [
        ("MADDPG", "bert_networks_multi_agent", "bert_networks_multi_agent"),
        ("MATD3", "bert_networks_multi_agent", "bert_matd3_critic_networks"),
    ],
)
@pytest.mark.parametrize(
    "observation_space, net_config", [("ma_vector_space", "encoder_mlp_config")]
)
@pytest.mark.parametrize("action_space", ["ma_discrete_space"])
@pytest.mark.parametrize("INIT_HP", [SHARED_INIT_HP_MA])
@pytest.mark.parametrize("population_size", [1])
@pytest.mark.parametrize("hp_config", [None])
@pytest.mark.parametrize("accelerator", [None, Accelerator(device_placement=False)])
@pytest.mark.parametrize("torch_compiler", [None])
def test_mutation_applies_bert_architecture_mutations_multi_agent(
    algo,
    device,
    accelerator,
    init_pop,
    observation_space,
    action_space,
    request,
    actor_network,
    critic_network,
):
    observation_space = request.getfixturevalue(observation_space)
    action_space = request.getfixturevalue(action_space)

    # Pass the network parameters to init_pop through the test
    actual_actor_network = (
        request.getfixturevalue(actor_network) if actor_network else None
    )
    actual_critic_network = (
        request.getfixturevalue(critic_network) if critic_network else None
    )

    # Create a custom population with the BERT networks
    from agilerl.utils.utils import create_population

    population = create_population(
        algo=algo,
        observation_space=observation_space,
        action_space=action_space,
        hp_config=None,
        net_config=request.getfixturevalue("encoder_mlp_config"),
        INIT_HP=SHARED_INIT_HP_MA,
        population_size=1,
        device=device,
        accelerator=accelerator,
        actor_network=actual_actor_network,
        critic_network=actual_critic_network,
    )

    mutations = Mutations(
        0,
        1,
        0.5,
        0,
        0,
        0,
        0.5,
        device=device,
        accelerator=accelerator,
    )

    test_agent = "agent_0"
    mut_methods = population[0].actors[test_agent].mutation_methods
    for mut_method in mut_methods:

        class DummyRNG:
            def choice(self, a, size=None, replace=True, p=None):
                return [".".join([test_agent, mut_method])]

        mutations.rng = DummyRNG()

        new_population = [agent.clone(wrap=False) for agent in population]
        mutated_population = [
            mutations.architecture_mutate(agent) for agent in new_population
        ]

        assert len(mutated_population) == len(population)
        for old, individual in zip(population, mutated_population):
            policy_name = individual.registry.policy()
            policy = getattr(individual, policy_name)
            # old_policy = getattr(old, policy_name)
            if policy.last_mutation_attr is not None:
                sampled_mutation = ".".join(policy.last_mutation_attr.split(".")[1:])
            else:
                sampled_mutation = None

            assert individual.mut == sampled_mutation or "None"

            if sampled_mutation is not None:
                for group in old.registry.groups:
                    if group.eval_network != policy_name:
                        eval_module = getattr(individual, group.eval_network)
                        # old_eval_module = getattr(old, group.eval_network)
                        for _, module in eval_module.items():
                            bottom_eval_mut = module.last_mutation_attr.split(".")[-1]
                            bottom_policy_mut = policy.last_mutation_attr.split(".")[-1]
                            assert module.last_mutation_attr is not None
                            assert bottom_eval_mut == bottom_policy_mut

            assert old.index == individual.index

        del new_population, mutated_population
    del mutations, population


@pytest.mark.parametrize(
    "use_accelerator, use_deepspeed_optimizer",
    [
        (True, True),
        (True, False),
        (False, False),
    ],
)
@pytest.mark.parametrize("algo", ["GRPO", "DPO"])
@pytest.mark.parametrize(
    "hp_to_mutate",
    [
        "lr",
        # "max_grad_norm"
    ],
)
def test_mutation_applies_rl_hp_mutation_llm_algorithm(
    request,
    vector_space,
    monkeypatch,
    use_accelerator,
    use_deepspeed_optimizer,
    algo,
    hp_to_mutate,
    grpo_hp_config,
    deepspeed_env,
):

    if hp_to_mutate == "lr":
        grpo_hp_config = grpo_hp_config
    elif hp_to_mutate == "max_grad_norm":
        grpo_hp_config = HyperparameterConfig(
            max_grad_norm=RLParameter(min=0.1, max=1.0),
        )

    pre_training_mut = False

    if use_accelerator:
        AcceleratorState._reset_state(True)

        deepspeed_config = {
            "gradient_accumulation_steps": 1,
            "zero_optimization": {
                "stage": 2,
            },
            "gradient_clipping": 0.3,
        }
        if use_deepspeed_optimizer:
            deepspeed_config["optimizer"] = {
                "type": "AdamW",
                "params": {
                    "lr": 1e-4,  # Smaller learning rate
                    "betas": [0.9, 0.999],
                    "eps": 1e-8,
                    "weight_decay": 0.01,
                },
            }
        accelerator = Accelerator(
            deepspeed_plugin=DeepSpeedPlugin(hf_ds_config=deepspeed_config),
        )
    else:
        accelerator = None
    init_hp = {
        "PAD_TOKEN_ID": 1000 - 1,
        "PAD_TOKEN": "<pad>",
        "BATCH_SIZE": 2,
        "BETA": 0.001,
        "LR": 0.001,
        "MAX_GRAD_NORM": 0.5,
        "UPDATE_EPOCHS": 1,
        "MAX_MODEL_LEN": 100,
    }
    device = "cuda" if torch.cuda.is_available() else "cpu"
    population = create_population(
        algo=algo,
        observation_space=vector_space,
        action_space=copy.deepcopy(vector_space),
        net_config=None,
        INIT_HP=init_hp,
        hp_config=grpo_hp_config,
        actor_network=create_module(
            input_size=10,
            max_tokens=20,
            vocab_size=1000,
            device=device,
        ),
        algo_kwargs={
            "lora_config": LoraConfig(
                r=16,
                lora_alpha=64,
                target_modules=["linear_1"],
                task_type="CAUSAL_LM",
                lora_dropout=0.05,
            ),
            "pad_token_id": 1000 - 1,
            "pad_token": "<pad>",
        },
        accelerator=accelerator,
        device=device,
    )

    mutations = Mutations(
        0,
        0,
        0,
        0,
        0,
        1,
        1,
        device=device,
        accelerator=accelerator,
    )

    print("original lr: ", [agent.lr for agent in population])

    new_population = [agent.clone(wrap=False) for agent in population]
    mutated_population = mutations.mutation(new_population, pre_training_mut)

    print("mutated lr: ", [agent.lr for agent in mutated_population])

    assert len(mutated_population) == len(population)
    for old, individual in zip(population, mutated_population):
        available_mutations = grpo_hp_config.names()
        assert individual.mut in available_mutations

        new_value = getattr(individual, individual.mut)
        min_value = grpo_hp_config[individual.mut].min
        max_value = grpo_hp_config[individual.mut].max
        assert min_value <= new_value <= max_value
        assert old.index == individual.index
    for agent in mutated_population:
        opt = (
            agent.actor.optimizer
            if (use_deepspeed_optimizer and use_accelerator)
            else agent.optimizer.optimizer
        )
        for param_group in opt.param_groups:
            assert param_group["lr"] == agent.lr
        if use_accelerator:
            assert (
                agent.accelerator.state.deepspeed_plugin.deepspeed_config[
                    "gradient_clipping"
                ]
                == agent.max_grad_norm
            )
    for mut_agent, old_agent in zip(mutated_population, new_population):
        mut_agent.clean_up()
        old_agent.clean_up()
    if use_accelerator:
        AcceleratorState._reset_state(True)


@pytest.mark.parametrize("mutation_type", ["architecture", "parameters", "activation"])
@pytest.mark.parametrize("algo", ["GRPO", "DPO"])
def test_mutations_warns_on_llm_algorithm(
    request, grpo_hp_config, vector_space, mutation_type, algo
):
    pre_training_mut = False
    init_hp = {
        "PAD_TOKEN_ID": 1000 - 1,
        "PAD_TOKEN": "<pad>",
        "BATCH_SIZE": 2,
        "BETA": 0.001,
        "LR": 5e-7,
        "MAX_GRAD_NORM": 0.1,
        "UPDATE_EPOCHS": 1,
        "MAX_MODEL_LEN": 100,
    }

    population = create_population(
        algo=algo,
        observation_space=vector_space,
        action_space=copy.deepcopy(vector_space),
        net_config=None,
        INIT_HP=init_hp,
        hp_config=grpo_hp_config,
        actor_network=create_module(
            input_size=10,
            max_tokens=20,
            vocab_size=1000,
            device="cuda" if torch.cuda.is_available() else "cpu",
        ),
        algo_kwargs={
            "lora_config": LoraConfig(
                r=16,
                lora_alpha=64,
                target_modules=["linear_1"],
                task_type="CAUSAL_LM",
                lora_dropout=0.05,
            ),
            "pad_token_id": 1000 - 1,
            "pad_token": "<pad>",
        },
    )

    mutations = Mutations(
        0,
        1 if mutation_type == "architecture" else 0,
        0.5 if mutation_type == "architecture" else 0,
        1 if mutation_type == "parameters" else 0,
        1 if mutation_type == "activation" else 0,
        0,
        0.1,
        device="cuda" if torch.cuda.is_available() else "cpu",
        accelerator=None,
    )

    new_population = [agent.clone(wrap=False) for agent in population]

    if mutation_type == "architecture":
        with pytest.raises(MutationError):
            mutations.mutation(new_population, pre_training_mut)

        # Since MutationError is expected, create a dummy mutated_population for the assertions
        mutated_population = new_population
        for individual in mutated_population:
            individual.mut = "None"
    else:
        with pytest.warns(UserWarning):
            mutated_population = mutations.mutation(new_population, pre_training_mut)

    assert len(mutated_population) == len(population)
    for old, individual in zip(population, mutated_population):
        assert old.mut is None
        assert individual.mut == "None"

    for mut_agent, old_agent in zip(mutated_population, new_population):
        mut_agent.clean_up()
        old_agent.clean_up()
    del mutations
    del population
    del mutated_population
    del new_population



from .bc_lm import BC_LM, BC_Evaluator, BC_Policy
from .cqn import CQN
from .ddpg import DDPG
from .dpo import DPO
from .dqn import DQN
from .dqn_rainbow import RainbowDQN
from .grpo import GRPO
from .ilql import ILQL
from .ippo import IPPO
from .maddpg import MADDPG
from .matd3 import MATD3
from .neural_ts_bandit import NeuralTS
from .neural_ucb_bandit import NeuralUCB
from .ppo import PPO
from .td3 import TD3

__all__ = [
    "BC_LM",
    "BC_Policy",
    "BC_Evaluator",
    "CQN",
    "DDPG",
    "DQN",
    "RainbowDQN",
    "ILQL",
    "IPPO",
    "MADDPG",
    "MATD3",
    "NeuralTS",
    "NeuralUCB",
    "PPO",
    "TD3",
    "GRPO",
    "DPO",
]



from collections import deque
from typing import Deque, Optional, Union

import torch
from tensordict import TensorDict, TensorDictBase, is_tensor_collection

from agilerl.components.segment_tree import MinSegmentTree, SumSegmentTree
from agilerl.typing import ArrayOrTensor

DataType = Union[dict[str, ArrayOrTensor], TensorDict]


class ReplayBuffer:
    """A circular replay buffer for off-policy learning using a TensorDict as storage.

    :param max_size: Maximum number of transitions to store
    :type max_size: int
    :param device: Device to store the transitions on
    :type device: Optional[Union[str, torch.device]], optional
    :param dtype: Data type for the tensors
    :type dtype: torch.dtype, optional
    """

    def __init__(
        self,
        max_size: int,
        device: Union[str, torch.device] = "cpu",
        dtype: torch.dtype = torch.float32,
    ) -> None:
        self.max_size = max_size
        self.device = device
        self.dtype = dtype
        self.counter = 0
        self.initialized = False

        self._cursor = 0
        self._size = 0
        self._storage: Optional[TensorDict] = None

    @property
    def storage(self) -> TensorDict:
        """Storage of the buffer."""
        return self._storage

    @property
    def size(self) -> int:
        """Number of transitions in the buffer."""
        return self._size

    @size.setter
    def size(self, value: int) -> None:
        self._size = value

    @property
    def is_full(self) -> bool:
        return len(self) == self.max_size

    def __len__(self) -> int:
        return self._size

    def _init(self, data: TensorDict) -> None:
        """Initialize the buffer given the passed data. For each key,
        we inspect the shape of the value and initialize the storage
        tensor with the correct shape.

        :param data: Data to initialize the buffer with
        :type data: TensorDict
        """
        _data: TensorDict = data[0]
        self._storage = torch.zeros_like(_data.expand((self.max_size, *_data.shape)))
        self.initialized = True

    def add(self, data: TensorDict) -> None:
        """Add a transition to the buffer.

        :param data: Transition to add to the buffer
        :type data: Union[TensorDict, dict[str, Any]]
        """
        # Initialize storage
        data = data.to(self.device)
        _n_transitions = data.shape[0]

        # Ensure all tensors in data have proper dimensions beyond batch dimension
        # Handles the case of scalar observations that become (batch_size,)
        # instead of (batch_size, 1)
        for key, value in data.items():
            if is_tensor_collection(value):
                value: TensorDictBase = value
                for k, v in value.items():
                    if v.ndim == 1:
                        value[k] = v.reshape(_n_transitions, 1)
            else:
                if value.ndim == 1:
                    value = value.reshape(_n_transitions, 1)

            data[key] = value

        if self._storage is None:
            self._init(data)

        # Add to circular storage
        start = self._cursor
        end = self._cursor + _n_transitions
        if end > self.max_size:
            n = self.max_size - start
            self._storage[start:] = data[:n]
            self._storage[: _n_transitions - n] = data[n:]
        else:
            self._storage[start:end] = data

        # Update cursor and size
        self._cursor = end % self.max_size
        self._size = min(self._size + _n_transitions, self.max_size)
        self.counter += _n_transitions

    def sample(self, batch_size: int, return_idx: bool = False) -> TensorDict:
        """Sample a batch of transitions.

        :param batch_size: Number of samples to return
        :type batch_size: int
        :param return_idx: Boolean flag to return index of samples randomly selected, defaults to False
        :type return_idx: bool, optional
        :return: TensorDict containing sampled experiences
        :rtype: TensorDict
        """
        # Ensure samples are unique
        indices = torch.randperm(self.size)[:batch_size]
        samples: TensorDict = self._storage[indices]

        if return_idx:
            samples["idxs"] = indices

        return samples

    def clear(self) -> None:
        """Clear all transitions from the buffer."""
        self._size = 0
        self._cursor = 0
        self._storage = None
        self.initialized = False


class MultiStepReplayBuffer(ReplayBuffer):
    """A circular replay buffer for n-step returns in off-policy learning.

    :param max_size: Maximum number of transitions to store
    :type max_size: int
    :param n_step: Number of steps to accumulate reward over
    :type n_step: int
    :param gamma: Discount factor
    :type gamma: float
    :param device: Device to store the transitions on
    :type device: Optional[Union[str, torch.device]], optional
    :param dtype: Data type for the tensors
    :type dtype: torch.dtype, optional
    """

    def __init__(
        self,
        max_size: int,
        n_step: int = 3,
        gamma: float = 0.99,
        device: Union[str, torch.device] = "cpu",
        dtype: torch.dtype = torch.float32,
    ) -> None:
        super().__init__(max_size, device, dtype)

        self.n_step = n_step
        self.gamma = gamma
        self.n_step_buffer: Deque[TensorDict] = deque(maxlen=n_step)
        self.reward_key = "reward"
        self.done_key = None
        self.ns_key = "next_obs"

    def add(self, data: TensorDict) -> Optional[TensorDict]:
        """Add a transition to the n-step buffer and potentially to the replay buffer.

        :param data: Transition to add to the buffer
        :type data: TensorDict
        :return: First transition in the n-step buffer
        :rtype: Optional[TensorDict]
        """
        # Add to n-step buffer
        data = data.to(self.device)
        self.n_step_buffer.append(data)

        # If buffer is not full yet, don't process n-step return
        if len(self.n_step_buffer) < self.n_step:
            return

        # Calculate n-step return
        n_step_data = self._get_n_step_info()

        # Add to replay buffer
        super().add(n_step_data)
        return self.n_step_buffer[0]

    def sample_from_indices(self, idxs: torch.Tensor) -> TensorDict:
        """Sample a batch of transitions from the buffer using the provided indices.

        :param idxs: Indices of the transitions to sample
        :type idxs: torch.Tensor
        :return: TensorDict containing sampled experiences
        :rtype: TensorDict
        """
        return self.storage[idxs]

    def _get_n_step_info(self) -> TensorDict:
        """Calculate the n-step return information.

        :return: Transition with n-step return
        :rtype: TensorDict
        """
        # Copy the first transition as a base
        first_transition: TensorDict = self.n_step_buffer[0].clone()

        # Get the reward key based on what's available in the transition
        if not self.initialized:
            assert (
                self.reward_key in self.n_step_buffer[0]
            ), f"Reward key not found in transition. Expected key: {self.reward_key}"
            assert (
                self.ns_key in self.n_step_buffer[0]
            ), f"Next observation key not found in transition. Expected key: {self.ns_key}"

            done_key = None
            expected_keys = ["done", "termination", "terminated"]
            for key in expected_keys:
                if key in self.n_step_buffer[0]:
                    done_key = key
                    break

            assert (
                done_key is not None
            ), f"No done/termination key found in transition. Expected keys: {expected_keys}"
            self.done_key = done_key

        # Start with reward from first transition
        n_step_reward: torch.Tensor = first_transition[self.reward_key]
        n_step_reward = n_step_reward.clone()

        # Get the last next_state and done flag
        for i, transition in enumerate(list(self.n_step_buffer)[1:]):
            # Add discounted reward
            reward: torch.Tensor = transition[self.reward_key]
            n_step_reward += reward * (self.gamma ** (i + 1))

            # Update next_state and done flag
            done: torch.Tensor = transition[self.done_key]
            next_obs: torch.Tensor = transition[self.ns_key]
            first_transition[self.ns_key] = next_obs.clone()
            first_transition[self.done_key] = done.clone()

            if done.bool().any():  # Stop if episode terminated
                break

        # Update the reward with n-step return
        first_transition[self.reward_key] = n_step_reward

        return first_transition


class PrioritizedReplayBuffer(ReplayBuffer):
    """A prioritized replay buffer for off-policy learning as introduced in the paper
    'Prioritized Experience Replay' (Schaul et al., 2015).

    :param max_size: Maximum number of transitions to store
    :type max_size: int
    :param alpha: How much prioritization to use (0 - no prioritization, 1 - full prioritization)
    :type alpha: float
    :param device: Device to store the transitions on.
    :type device: Optional[Union[str, torch.device]], optional
    :param dtype: Data type for the tensors
    :type dtype: torch.dtype, optional
    """

    def __init__(
        self,
        max_size: int,
        alpha: float = 0.6,
        device: Union[str, torch.device] = "cpu",
        dtype: torch.dtype = torch.float32,
    ) -> None:
        super().__init__(max_size, device, dtype)
        self.alpha = alpha
        self.max_priority = 1.0
        self.tree_ptr = 0

        # Find the closest power of 2 capacity for the segment trees
        tree_capacity = 1
        while tree_capacity < max_size:
            tree_capacity *= 2

        # Initialize segment trees
        self.sum_tree = SumSegmentTree(tree_capacity)
        self.min_tree = MinSegmentTree(tree_capacity)

    def add(self, data: TensorDict) -> None:
        """Add a transition to the buffer.

        :param data: Transition to add to the buffer
        :type data: TensorDict
        """
        # Add to replay buffer
        super().add(data)

        # Add max priority for new entries
        n_transitions = data.shape[0]
        for i in range(n_transitions):
            self._update_priority(self.tree_ptr, self.max_priority)
            self.tree_ptr = (self.tree_ptr + 1) % self.max_size

    def _update_priority(self, idx: int, priority: float) -> None:
        """Update the priority of an experience in the buffer.

        :param idx: Index of the experience
        :type idx: int
        :param priority: New priority value
        :type priority: float
        """
        assert 0 <= idx < self.max_size

        # Apply alpha to priority
        priority_alpha = priority**self.alpha

        # Update trees
        self.sum_tree[idx] = priority_alpha
        self.min_tree[idx] = priority_alpha

        # Update max priority
        self.max_priority = max(self.max_priority, priority)

    def sample(self, batch_size: int, beta: float = 0.4) -> TensorDict:
        """Sample a batch of transitions based on priorities.

        :param batch_size: Number of samples to return
        :type batch_size: int
        :param beta: Beta parameter for importance sampling, defaults to 0.4
        :type beta: float, optional
        :return: Batch of transitions
        :rtype: TensorDict
        """
        # Sample indices based on priorities
        indices = self._sample_proportional(batch_size)

        # Gather transitions
        samples: TensorDict = self.storage[indices]
        samples = samples.clone()

        # Calculate importance sampling weights
        weights = self._calculate_weights(indices, beta)

        # Add weights and indices to the batch
        samples["weights"] = weights.unsqueeze(1)
        samples["idxs"] = indices.unsqueeze(1)

        return samples

    def _sample_proportional(self, batch_size: int) -> torch.Tensor:
        """Sample indices based on their priorities.

        :param batch_size: Number of samples
        :type batch_size: int
        :return: Sampled indices
        :rtype: torch.Tensor
        """
        indices = torch.zeros(batch_size, dtype=torch.int64)

        # Get the sum of all priorities and section length
        total_priority = self.sum_tree.sum()
        segment = total_priority / batch_size

        # Sample from each segment
        for i in range(batch_size):
            a = segment * i
            b = segment * (i + 1)

            # Sample uniformly from segment and retrieve corresponding index
            upperbound = torch.rand(1).item() * (b - a) + a
            idx = self.sum_tree.retrieve(upperbound)
            indices[i] = idx

        return indices

    def _calculate_weights(self, indices: torch.Tensor, beta: float) -> torch.Tensor:
        """Calculate importance sampling weights for prioritized replay.

        :param indices: Sampled indices
        :type indices: torch.Tensor
        :param beta: Beta parameter for importance sampling
        :type beta: float
        :return: Weights for the sampled transitions
        :rtype: torch.Tensor
        """
        # Create a tensor for weights
        batch_size = len(indices)
        weights = torch.zeros(batch_size, device=self.device)

        # Find the min probability from the min tree
        p_min = self.min_tree.min() / self.sum_tree.sum()

        # Calculate the max weight value
        max_weight = (p_min * self.size) ** -beta

        # Calculate weights for each index
        for i, idx in enumerate(indices):
            p_sample = self.sum_tree[idx] / self.sum_tree.sum()
            weight = (p_sample * self.size) ** -beta
            weights[i] = weight / max_weight  # Normalize

        return weights

    def update_priorities(
        self, indices: torch.Tensor, priorities: torch.Tensor
    ) -> None:
        """Update priorities of the sampled transitions.

        :param indices: Indices of transitions to update
        :type indices: torch.Tensor
        :param priorities: New priorities
        :type priorities: torch.Tensor
        """
        for idx, priority in zip(indices, priorities):
            # Handle small priorities
            priority = max(priority.item(), 1e-5)

            # Update the priority
            self._update_priority(idx.item(), priority)



import h5py
from accelerate import Accelerator

from agilerl.components.replay_buffer import ReplayBuffer
from agilerl.hpo.mutation import Mutations
from agilerl.hpo.tournament import TournamentSelection
from agilerl.training.train_offline import train_offline
from agilerl.utils.utils import (
    create_population,
    make_vect_envs,
    observation_space_channels_to_first,
    print_hyperparams,
)

# !Note: If you are running this demo without having installed agilerl,
# uncomment and place the following above agilerl imports:

# import sys
# sys.path.append('../')


def main(INIT_HP, MUTATION_PARAMS, NET_CONFIG):
    accelerator = Accelerator()

    accelerator.wait_for_everyone()
    if accelerator.is_main_process:
        print("============ AgileRL Distributed ============")
    accelerator.wait_for_everyone()

    env = make_vect_envs(INIT_HP["ENV_NAME"], num_envs=INIT_HP["NUM_ENVS"])

    observation_space = env.single_observation_space
    action_space = env.single_action_space
    if INIT_HP["CHANNELS_LAST"]:
        observation_space = observation_space_channels_to_first(observation_space)

    dataset = h5py.File(INIT_HP["DATASET"], "r")

    memory = ReplayBuffer(INIT_HP["MEMORY_SIZE"], device=accelerator.device)
    tournament = TournamentSelection(
        INIT_HP["TOURN_SIZE"],
        INIT_HP["ELITISM"],
        INIT_HP["POP_SIZE"],
        INIT_HP["EVAL_LOOP"],
    )
    mutations = Mutations(
        no_mutation=MUTATION_PARAMS["NO_MUT"],
        architecture=MUTATION_PARAMS["ARCH_MUT"],
        new_layer_prob=MUTATION_PARAMS["NEW_LAYER"],
        parameters=MUTATION_PARAMS["PARAMS_MUT"],
        activation=MUTATION_PARAMS["ACT_MUT"],
        rl_hp=MUTATION_PARAMS["RL_HP_MUT"],
        mutation_sd=MUTATION_PARAMS["MUT_SD"],
        min_lr=MUTATION_PARAMS["MIN_LR"],
        max_lr=MUTATION_PARAMS["MAX_LR"],
        min_batch_size=MUTATION_PARAMS["MAX_BATCH_SIZE"],
        max_batch_size=MUTATION_PARAMS["MAX_BATCH_SIZE"],
        rand_seed=MUTATION_PARAMS["RAND_SEED"],
        accelerator=accelerator,
    )

    agent_pop = create_population(
        algo=INIT_HP["ALGO"],
        observation_space=observation_space,
        action_space=action_space,
        net_config=NET_CONFIG,
        INIT_HP=INIT_HP,
        population_size=INIT_HP["POP_SIZE"],
        num_envs=INIT_HP["NUM_ENVS"],
        accelerator=accelerator,
    )

    trained_pop, pop_fitnesses = train_offline(
        env,
        INIT_HP["ENV_NAME"],
        dataset,
        INIT_HP["ALGO"],
        agent_pop,
        memory=memory,
        INIT_HP=INIT_HP,
        MUT_P=MUTATION_PARAMS,
        swap_channels=INIT_HP["CHANNELS_LAST"],
        max_steps=INIT_HP["MAX_STEPS"],
        evo_steps=INIT_HP["EVO_STEPS"],
        eval_steps=INIT_HP["EVAL_STEPS"],
        eval_loop=INIT_HP["EVAL_LOOP"],
        target=INIT_HP["TARGET_SCORE"],
        tournament=tournament,
        mutation=mutations,
        wb=INIT_HP["WANDB"],
        accelerator=accelerator,
    )

    print_hyperparams(trained_pop)
    # plot_population_score(trained_pop)

    env.close()


if __name__ == "__main__":
    INIT_HP = {
        "ENV_NAME": "CartPole-v1",  # Gym environment name
        "DATASET": "data/cartpole/cartpole_random_v1.1.0.h5",  # Offline RL dataset
        "ALGO": "CQN",  # Algorithm
        "DOUBLE": True,  # Use double Q-learning
        # Swap image channels dimension from last to first [H, W, C] -> [C, H, W]
        "CHANNELS_LAST": False,
        "NUM_ENVS": 16,
        "BATCH_SIZE": 256,  # Batch size
        "LR": 1e-3,  # Learning rate
        "MAX_STEPS": 1_000_000,  # Max no. steps
        "TARGET_SCORE": 200.0,  # Early training stop at avg score of last 100 episodes
        "GAMMA": 0.99,  # Discount factor
        "MEMORY_SIZE": 10000,  # Max memory buffer size
        "LEARN_STEP": 1,  # Learning frequency
        "TAU": 1e-3,  # For soft update of target parameters
        "TOURN_SIZE": 2,  # Tournament size
        "ELITISM": True,  # Elitism in tournament selection
        "POP_SIZE": 6,  # Population size
        "EVO_STEPS": 10_000,  # Evolution frequency
        "EVAL_STEPS": None,  # Evaluation steps
        "EVAL_LOOP": 1,  # Evaluation episodes
        "WANDB": True,  # Log with Weights and Biases
    }

    MUTATION_PARAMS = {  # Relative probabilities
        "NO_MUT": 0.4,  # No mutation
        "ARCH_MUT": 0.2,  # Architecture mutation
        "NEW_LAYER": 0.2,  # New layer mutation
        "PARAMS_MUT": 0.2,  # Network parameters mutation
        "ACT_MUT": 0,  # Activation layer mutation
        "RL_HP_MUT": 0.2,  # Learning HP mutation
        # Learning HPs to choose from
        "RL_HP_SELECTION": ["lr", "batch_size"],
        "MUT_SD": 0.1,  # Mutation strength
        "RAND_SEED": 1,  # Random seed
    }

    NET_CONFIG = {
        "encoder_config": {
            "hidden_size": [32, 32],  # Actor hidden size
        }
    }

    main(INIT_HP, MUTATION_PARAMS, NET_CONFIG)



import copy

import numpy as np
import pytest
import torch
import torch.nn as nn
import torch.optim as optim
from accelerate import Accelerator
from accelerate.optimizer import AcceleratedOptimizer
from gymnasium import spaces

from agilerl.algorithms.ddpg import DDPG
from agilerl.modules import EvolvableCNN, EvolvableMLP, EvolvableMultiInput
from agilerl.networks.actors import DeterministicActor
from agilerl.networks.q_networks import ContinuousQNetwork
from agilerl.wrappers.make_evolvable import MakeEvolvable
from tests.helper_functions import (
    assert_not_equal_state_dict,
    assert_state_dicts_equal,
    get_experiences_batch,
    get_sample_from_space,
)


class DummyDDPG(DDPG):
    def __init__(self, observation_space, action_space, *args, **kwargs):
        super().__init__(observation_space, action_space, *args, **kwargs)

        self.tensor_test = torch.randn(1)


class DummyEnv:
    def __init__(self, state_size, vect=True, num_envs=2):
        self.state_size = state_size
        self.vect = vect
        if self.vect:
            self.state_size = (num_envs,) + self.state_size
            self.n_envs = num_envs
            self.num_envs = num_envs
        else:
            self.n_envs = 1

    def reset(self):
        return np.random.rand(*self.state_size), {}

    def step(self, action):
        return (
            np.random.rand(*self.state_size),
            np.random.randint(0, 5, self.n_envs),
            np.random.randint(0, 2, self.n_envs),
            np.random.randint(0, 2, self.n_envs),
            {},
        )


class SimpleCNN(nn.Module):
    def __init__(self):
        super().__init__()

        self.conv1 = nn.Conv2d(
            3, 16, kernel_size=3, stride=1, padding=1
        )  # Input channels: 3 (for RGB images), Output channels: 16
        self.relu1 = nn.ReLU()
        self.mp1 = nn.MaxPool2d(kernel_size=2, stride=2)
        self.conv2 = nn.Conv2d(
            16, 32, kernel_size=3, stride=1, padding=1
        )  # Input channels: 16, Output channels: 32
        self.relu2 = nn.ReLU()
        self.mp2 = nn.MaxPool2d(kernel_size=2, stride=2)
        self.flat = nn.Flatten()  # Flatten the 2D feature map to a 1D vector
        self.linear1 = nn.Linear(
            32 * 8 * 8, 128
        )  # Fully connected layer with 128 output features
        self.relu3 = nn.ReLU()
        self.linear2 = nn.Linear(
            128, 128
        )  # Fully connected layer with 128 output features

    def forward(self, x, xc):
        x = self.mp1(self.relu1(self.conv1(x)))
        x = self.mp2(self.relu2(self.conv2(x)))
        x = self.flat(x)
        x = self.relu3(self.linear1(x))
        x = self.relu3(self.linear2(x))
        return x


# initialize ddpg with valid parameters
@pytest.mark.parametrize(
    "observation_space, encoder_cls",
    [
        ("vector_space", EvolvableMLP),
        ("image_space", EvolvableCNN),
        ("dict_space", EvolvableMultiInput),
        ("multidiscrete_space", EvolvableMLP),
    ],
)
@pytest.mark.parametrize("accelerator", [None, Accelerator()])
def test_initialize_ddpg(observation_space, encoder_cls, accelerator, request):
    action_space = spaces.Box(low=-1, high=1, shape=(2,), dtype=np.float32)
    observation_space = request.getfixturevalue(observation_space)
    ddpg = DDPG(observation_space, action_space, accelerator=accelerator)

    expected_opt_cls = AcceleratedOptimizer if accelerator else optim.Adam
    expected_device = accelerator.device if accelerator else "cpu"
    assert ddpg.observation_space == observation_space
    assert ddpg.action_space == action_space
    assert ddpg.batch_size == 64
    assert ddpg.lr_actor == 0.0001
    assert ddpg.lr_critic == 0.001
    assert ddpg.learn_step == 5
    assert ddpg.gamma == 0.99
    assert ddpg.tau == 0.001
    assert ddpg.mut is None
    assert ddpg.device == expected_device
    assert ddpg.accelerator == accelerator
    assert ddpg.index == 0
    assert ddpg.scores == []
    assert ddpg.fitness == []
    assert ddpg.steps == [0]
    assert isinstance(ddpg.actor.encoder, encoder_cls)
    assert isinstance(ddpg.actor_target.encoder, encoder_cls)
    assert isinstance(ddpg.actor_optimizer.optimizer, expected_opt_cls)
    assert isinstance(ddpg.critic.encoder, encoder_cls)
    assert isinstance(ddpg.critic_target.encoder, encoder_cls)
    assert isinstance(ddpg.critic_optimizer.optimizer, expected_opt_cls)
    assert isinstance(ddpg.criterion, nn.MSELoss)
    ddpg.clean_up()


# Can initialize ddpg with an actor network
# TODO: This will be deprecated in the future
@pytest.mark.parametrize(
    "observation_space, actor_network, critic_network, input_tensor, input_tensor_critic",
    [
        (
            "vector_space",
            "simple_mlp",
            "simple_mlp",
            torch.randn(1, 4),
            torch.randn(1, 4),
        ),
    ],
)
def test_initialize_ddpg_with_actor_network(
    observation_space,
    actor_network,
    critic_network,
    input_tensor,
    input_tensor_critic,
    request,
):
    action_space = spaces.Box(low=-1, high=1, shape=(2,), dtype=np.float32)
    observation_space = request.getfixturevalue(observation_space)
    actor_network = request.getfixturevalue(actor_network)
    actor_network = MakeEvolvable(actor_network, input_tensor)
    critic_network = request.getfixturevalue(critic_network)
    critic_network = MakeEvolvable(critic_network, input_tensor_critic)

    ddpg = DDPG(
        observation_space,
        action_space,
        actor_network=actor_network,
        critic_network=critic_network,
    )

    assert ddpg.observation_space == observation_space
    assert ddpg.action_space == action_space
    assert ddpg.batch_size == 64
    assert ddpg.lr_actor == 0.0001
    assert ddpg.lr_critic == 0.001
    assert ddpg.learn_step == 5
    assert ddpg.gamma == 0.99
    assert ddpg.tau == 0.001
    assert ddpg.mut is None
    assert ddpg.device == "cpu"
    assert ddpg.accelerator is None
    assert ddpg.index == 0
    assert ddpg.scores == []
    assert ddpg.fitness == []
    assert ddpg.steps == [0]
    assert isinstance(ddpg.actor_optimizer.optimizer, optim.Adam)
    assert isinstance(ddpg.critic_optimizer.optimizer, optim.Adam)
    assert isinstance(ddpg.criterion, nn.MSELoss)
    ddpg.clean_up()


def test_initialize_ddpg_with_actor_network_evo_net(vector_space):
    action_space = spaces.Box(low=-1, high=1, shape=(2,), dtype=np.float32)

    actor_network = DeterministicActor(vector_space, action_space)
    critic_network = ContinuousQNetwork(vector_space, action_space)

    ddpg = DDPG(
        vector_space,
        action_space,
        actor_network=actor_network,
        critic_network=critic_network,
    )

    assert ddpg.observation_space == vector_space
    assert ddpg.action_space == action_space
    assert ddpg.batch_size == 64
    assert ddpg.lr_actor == 0.0001
    assert ddpg.lr_critic == 0.001
    assert ddpg.learn_step == 5
    assert ddpg.gamma == 0.99
    assert ddpg.tau == 0.001
    assert ddpg.mut is None
    assert ddpg.device == "cpu"
    assert ddpg.accelerator is None
    assert ddpg.index == 0
    assert ddpg.scores == []
    assert ddpg.fitness == []
    assert ddpg.steps == [0]
    assert isinstance(ddpg.actor_optimizer.optimizer, optim.Adam)
    assert isinstance(ddpg.critic_optimizer.optimizer, optim.Adam)
    assert isinstance(ddpg.criterion, nn.MSELoss)
    ddpg.clean_up()
    ddpg.clean_up()


def test_initialize_ddpg_with_incorrect_actor_net(vector_space):
    action_space = spaces.Box(low=-1, high=1, shape=(2,), dtype=np.float32)
    actor_network = "dummy"
    critic_network = "dummy"
    with pytest.raises(TypeError):
        ddpg = DDPG(
            vector_space,
            action_space,
            expl_noise=np.zeros((1, action_space.shape[0])),
            actor_network=actor_network,
            critic_network=critic_network,
        )
        assert ddpg


# Can initialize ddpg with an actor network but no critic - should trigger warning
@pytest.mark.parametrize(
    "observation_space, actor_network, critic_network, input_tensor, input_tensor_critic",
    [
        (
            "vector_space",
            "simple_mlp",
            "simple_mlp_critic",
            torch.randn(1, 4),
            torch.randn(1, 6),
        ),
    ],
)
def test_initialize_ddpg_with_actor_network_no_critic(
    observation_space,
    actor_network,
    critic_network,
    input_tensor,
    input_tensor_critic,
    request,
):
    action_space = spaces.Box(low=-1, high=1, shape=(2,), dtype=np.float32)
    observation_space = request.getfixturevalue(observation_space)
    actor_network = request.getfixturevalue(actor_network)
    actor_network = MakeEvolvable(actor_network, input_tensor)

    ddpg = DDPG(
        observation_space,
        action_space,
        actor_network=actor_network,
        critic_network=None,
    )

    assert ddpg.observation_space == observation_space
    assert ddpg.action_space == action_space
    assert ddpg.batch_size == 64
    assert ddpg.lr_actor == 0.0001
    assert ddpg.lr_critic == 0.001
    assert ddpg.learn_step == 5
    assert ddpg.gamma == 0.99
    assert ddpg.tau == 0.001
    assert ddpg.mut is None
    assert ddpg.device == "cpu"
    assert ddpg.accelerator is None
    assert ddpg.index == 0
    assert ddpg.scores == []
    assert ddpg.fitness == []
    assert ddpg.steps == [0]
    assert ddpg.actor != actor_network
    # assert ddpg.critic_network is None
    assert isinstance(ddpg.actor_optimizer.optimizer, optim.Adam)
    assert isinstance(ddpg.critic_optimizer.optimizer, optim.Adam)
    assert isinstance(ddpg.criterion, nn.MSELoss)
    ddpg.clean_up()


@pytest.mark.parametrize(
    "observation_space",
    [
        "vector_space",
        "discrete_space",
        "image_space",
        "dict_space",
        "multidiscrete_space",
    ],
)
@pytest.mark.parametrize("action_dtype", [np.float32, np.float64])
def test_returns_expected_action_training(observation_space, request, action_dtype):
    accelerator = Accelerator()
    action_space = spaces.Box(low=-1, high=1, shape=(2,), dtype=action_dtype)
    observation_space = request.getfixturevalue(observation_space)

    # Test without accelerator
    ddpg = DDPG(observation_space, action_space)
    state = get_sample_from_space(observation_space)

    training = False
    action = ddpg.get_action(state, training)[0]
    assert len(action) == action_space.shape[0]
    for act in action:
        assert isinstance(act, np.float32)
        assert -1 <= act <= 1
    ddpg.clean_up()

    # Test with accelerator
    ddpg = DDPG(observation_space, action_space, accelerator=accelerator)
    state = get_sample_from_space(observation_space)
    training = True
    action = ddpg.get_action(state, training)[0]

    assert len(action) == action_space.shape[0]
    for act in action:
        assert isinstance(act, np.float32)
        assert -1 <= act <= 1
    ddpg.clean_up()

    # Test without OU noise
    ddpg = DDPG(
        observation_space, action_space, O_U_noise=False, accelerator=accelerator
    )
    state = get_sample_from_space(observation_space)
    training = True
    action = ddpg.get_action(state, training)[0]

    assert len(action) == action_space.shape[0]
    for act in action:
        assert isinstance(act, np.float32)
        assert -1 <= act <= 1
    ddpg.clean_up()


# learns from experiences and updates network parameters
@pytest.mark.parametrize(
    "observation_space",
    [
        "vector_space",
        "discrete_space",
        "image_space",
        "dict_space",
        "multidiscrete_space",
    ],
)
@pytest.mark.parametrize("accelerator", [None, Accelerator()])
def test_learns_from_experiences(observation_space, accelerator, request):
    action_space = spaces.Box(low=-1, high=1, shape=(2,), dtype=np.float32)
    observation_space = request.getfixturevalue(observation_space)
    batch_size = 4
    policy_freq = 4

    # Create an instance of the ddpg class
    ddpg = DDPG(
        observation_space,
        action_space,
        batch_size=batch_size,
        policy_freq=policy_freq,
        accelerator=accelerator,
    )

    # Copy state dict before learning - should be different to after updating weights
    actor = ddpg.actor
    actor_target = ddpg.actor_target
    actor_pre_learn_sd = copy.deepcopy(ddpg.actor.state_dict())
    critic = ddpg.critic
    critic_target = ddpg.critic_target
    critic_pre_learn_sd = copy.deepcopy(ddpg.critic.state_dict())

    for i in range(policy_freq * 2):
        # Create a batch of experiences & learn
        device = accelerator.device if accelerator else "cpu"
        experiences = get_experiences_batch(
            observation_space, action_space, batch_size, device
        )
        ddpg.scores.append(0)
        actor_loss, critic_loss = ddpg.learn(experiences)

    assert isinstance(actor_loss, float)
    assert isinstance(critic_loss, float)
    assert critic_loss >= 0.0
    assert actor == ddpg.actor
    assert actor_target == ddpg.actor_target
    assert_not_equal_state_dict(actor_pre_learn_sd, ddpg.actor.state_dict())
    assert critic == ddpg.critic
    assert critic_target == ddpg.critic_target
    assert_not_equal_state_dict(critic_pre_learn_sd, ddpg.critic.state_dict())
    ddpg.clean_up()


# Updates target network parameters with soft update
def test_soft_update():
    observation_space = spaces.Box(low=-1, high=1, shape=(4,), dtype=np.float32)
    action_space = spaces.Box(low=-1, high=1, shape=(2,), dtype=np.float32)
    net_config = {"encoder_config": {"hidden_size": [64, 64]}}
    batch_size = 64
    lr_actor = 1e-4
    lr_critic = 1e-3
    learn_step = 5
    gamma = 0.99
    tau = 1e-3
    mut = None
    actor_network = None
    device = "cpu"
    accelerator = None
    wrap = True

    ddpg = DDPG(
        observation_space,
        action_space,
        net_config=net_config,
        batch_size=batch_size,
        lr_actor=lr_actor,
        lr_critic=lr_critic,
        learn_step=learn_step,
        gamma=gamma,
        tau=tau,
        mut=mut,
        actor_network=actor_network,
        device=device,
        accelerator=accelerator,
        wrap=wrap,
    )

    ddpg.soft_update(ddpg.actor, ddpg.actor_target)

    eval_params = list(ddpg.actor.parameters())
    target_params = list(ddpg.actor_target.parameters())
    expected_params = [
        ddpg.tau * eval_param + (1.0 - ddpg.tau) * target_param
        for eval_param, target_param in zip(eval_params, target_params)
    ]

    assert all(
        torch.allclose(expected_param, target_param)
        for expected_param, target_param in zip(expected_params, target_params)
    )

    ddpg.soft_update(ddpg.critic, ddpg.critic_target)

    eval_params = list(ddpg.critic.parameters())
    target_params = list(ddpg.critic_target.parameters())
    expected_params = [
        ddpg.tau * eval_param + (1.0 - ddpg.tau) * target_param
        for eval_param, target_param in zip(eval_params, target_params)
    ]

    assert all(
        torch.allclose(expected_param, target_param)
        for expected_param, target_param in zip(expected_params, target_params)
    )
    ddpg.clean_up()


# Runs algorithm test loop
@pytest.mark.parametrize("observation_space", ["vector_space", "image_space"])
@pytest.mark.parametrize("num_envs", [1, 3])
def test_algorithm_test_loop(observation_space, num_envs, request):
    action_space = spaces.Box(low=-1, high=1, shape=(2,), dtype=np.float32)
    observation_space = request.getfixturevalue(observation_space)

    vect = num_envs > 1
    env = DummyEnv(state_size=observation_space.shape, vect=vect, num_envs=num_envs)

    # env = make_vect_envs("CartPole-v1", num_envs=num_envs)
    agent = DDPG(observation_space=observation_space, action_space=action_space)
    mean_score = agent.test(env, max_steps=10)
    assert isinstance(mean_score, float)
    agent.clean_up()


# Clones the agent and returns an identical agent.
@pytest.mark.parametrize("observation_space", ["vector_space"])
def test_clone_returns_identical_agent(observation_space, request):
    action_space = spaces.Box(low=-1, high=1, shape=(2,), dtype=np.float32)
    observation_space = request.getfixturevalue(observation_space)

    ddpg = DDPG(observation_space, action_space)
    ddpg.fitness = [200, 200, 200]
    ddpg.scores = [94, 94, 94]
    ddpg.steps = [2500]
    ddpg.tensor_attribute = torch.randn(1)
    clone_agent = ddpg.clone()

    assert clone_agent.observation_space == ddpg.observation_space
    assert clone_agent.action_space == ddpg.action_space
    assert clone_agent.batch_size == ddpg.batch_size
    assert clone_agent.lr_actor == ddpg.lr_actor
    assert clone_agent.lr_critic == ddpg.lr_critic
    assert clone_agent.learn_step == ddpg.learn_step
    assert clone_agent.gamma == ddpg.gamma
    assert clone_agent.tau == ddpg.tau
    assert clone_agent.mut == ddpg.mut
    assert clone_agent.device == ddpg.device
    assert clone_agent.accelerator == ddpg.accelerator
    assert_state_dicts_equal(clone_agent.actor.state_dict(), ddpg.actor.state_dict())
    assert_state_dicts_equal(
        clone_agent.actor_target.state_dict(), ddpg.actor_target.state_dict()
    )
    assert_state_dicts_equal(clone_agent.critic.state_dict(), ddpg.critic.state_dict())
    assert_state_dicts_equal(
        clone_agent.critic_target.state_dict(), ddpg.critic_target.state_dict()
    )
    assert_state_dicts_equal(
        clone_agent.actor_optimizer.state_dict(), ddpg.actor_optimizer.state_dict()
    )
    assert_state_dicts_equal(
        clone_agent.critic_optimizer.state_dict(), ddpg.critic_optimizer.state_dict()
    )
    assert clone_agent.fitness == ddpg.fitness
    assert clone_agent.steps == ddpg.steps
    assert clone_agent.scores == ddpg.scores
    assert clone_agent.tensor_attribute == ddpg.tensor_attribute
    ddpg.clean_up()
    clone_agent.clean_up()

    accelerator = Accelerator()
    ddpg = DDPG(observation_space, action_space, accelerator=accelerator)
    clone_agent = ddpg.clone()

    assert clone_agent.observation_space == ddpg.observation_space
    assert clone_agent.action_space == ddpg.action_space
    assert clone_agent.batch_size == ddpg.batch_size
    assert clone_agent.lr_actor == ddpg.lr_actor
    assert clone_agent.lr_critic == ddpg.lr_critic
    assert clone_agent.learn_step == ddpg.learn_step
    assert clone_agent.gamma == ddpg.gamma
    assert clone_agent.tau == ddpg.tau
    assert clone_agent.mut == ddpg.mut
    assert clone_agent.device == ddpg.device
    assert clone_agent.accelerator == ddpg.accelerator
    assert_state_dicts_equal(clone_agent.actor.state_dict(), ddpg.actor.state_dict())
    assert_state_dicts_equal(
        clone_agent.actor_target.state_dict(), ddpg.actor_target.state_dict()
    )
    assert_state_dicts_equal(clone_agent.critic.state_dict(), ddpg.critic.state_dict())
    assert_state_dicts_equal(
        clone_agent.critic_target.state_dict(), ddpg.critic_target.state_dict()
    )
    assert_state_dicts_equal(
        clone_agent.actor_optimizer.state_dict(), ddpg.actor_optimizer.state_dict()
    )
    assert_state_dicts_equal(
        clone_agent.critic_optimizer.state_dict(), ddpg.critic_optimizer.state_dict()
    )
    assert clone_agent.fitness == ddpg.fitness
    assert clone_agent.steps == ddpg.steps
    assert clone_agent.scores == ddpg.scores
    ddpg.clean_up()
    clone_agent.clean_up()

    accelerator = Accelerator()
    ddpg = DDPG(observation_space, action_space, accelerator=accelerator, wrap=False)
    clone_agent = ddpg.clone(wrap=False)

    assert clone_agent.observation_space == ddpg.observation_space
    assert clone_agent.action_space == ddpg.action_space
    assert clone_agent.batch_size == ddpg.batch_size
    assert clone_agent.lr_actor == ddpg.lr_actor
    assert clone_agent.lr_critic == ddpg.lr_critic
    assert clone_agent.learn_step == ddpg.learn_step
    assert clone_agent.gamma == ddpg.gamma
    assert clone_agent.tau == ddpg.tau
    assert clone_agent.mut == ddpg.mut
    assert clone_agent.device == ddpg.device
    assert clone_agent.accelerator == ddpg.accelerator
    assert_state_dicts_equal(clone_agent.actor.state_dict(), ddpg.actor.state_dict())
    assert_state_dicts_equal(
        clone_agent.actor_target.state_dict(), ddpg.actor_target.state_dict()
    )
    assert_state_dicts_equal(clone_agent.critic.state_dict(), ddpg.critic.state_dict())
    assert_state_dicts_equal(
        clone_agent.critic_target.state_dict(), ddpg.critic_target.state_dict()
    )
    assert_state_dicts_equal(
        clone_agent.actor_optimizer.state_dict(), ddpg.actor_optimizer.state_dict()
    )
    assert_state_dicts_equal(
        clone_agent.critic_optimizer.state_dict(), ddpg.critic_optimizer.state_dict()
    )
    assert clone_agent.fitness == ddpg.fitness
    assert clone_agent.steps == ddpg.steps
    assert clone_agent.scores == ddpg.scores
    ddpg.clean_up()
    clone_agent.clean_up()


def test_clone_new_index():
    observation_space = spaces.Box(low=-1, high=1, shape=(4,), dtype=np.float32)
    action_space = spaces.Box(low=-1, high=1, shape=(2,), dtype=np.float32)

    ddpg = DDPG(observation_space, action_space)
    clone_agent = ddpg.clone(index=100)

    assert clone_agent.index == 100
    ddpg.clean_up()
    clone_agent.clean_up()


def test_clone_after_learning():
    observation_space = spaces.Box(low=-1, high=1, shape=(4,), dtype=np.float32)
    action_space = spaces.Box(low=-1, high=1, shape=(2,), dtype=np.float32)

    batch_size = 8
    ddpg = DDPG(observation_space, action_space)

    experiences = get_experiences_batch(observation_space, action_space, batch_size)

    ddpg.learn(experiences)
    clone_agent = ddpg.clone()

    assert clone_agent.observation_space == ddpg.observation_space
    assert clone_agent.action_space == ddpg.action_space
    assert clone_agent.batch_size == ddpg.batch_size
    assert clone_agent.lr_actor == ddpg.lr_actor
    assert clone_agent.lr_critic == ddpg.lr_critic
    assert clone_agent.learn_step == ddpg.learn_step
    assert clone_agent.gamma == ddpg.gamma
    assert clone_agent.tau == ddpg.tau
    assert clone_agent.mut == ddpg.mut
    assert clone_agent.device == ddpg.device
    assert clone_agent.accelerator == ddpg.accelerator
    assert_state_dicts_equal(clone_agent.actor.state_dict(), ddpg.actor.state_dict())
    assert_state_dicts_equal(
        clone_agent.actor_target.state_dict(), ddpg.actor_target.state_dict()
    )
    assert_state_dicts_equal(clone_agent.critic.state_dict(), ddpg.critic.state_dict())
    assert_state_dicts_equal(
        clone_agent.critic_target.state_dict(), ddpg.critic_target.state_dict()
    )
    assert_state_dicts_equal(
        clone_agent.actor_optimizer.state_dict(), ddpg.actor_optimizer.state_dict()
    )
    assert_state_dicts_equal(
        clone_agent.critic_optimizer.state_dict(), ddpg.critic_optimizer.state_dict()
    )
    assert clone_agent.fitness == ddpg.fitness
    assert clone_agent.steps == ddpg.steps
    assert clone_agent.scores == ddpg.scores
    ddpg.clean_up()
    clone_agent.clean_up()


@pytest.mark.parametrize(
    "min, max, action, expected_result, device",
    [
        (0, 1, [1.1, 0.75, -1], [1.0, 0.75, 0], "cpu"),
        ([0.5, 0, 0.1], 1, [0, 0, 0.2], [0.5, 0, 0.2], "cpu"),
        (0, [0.75, 1.0, 0.1], [1.0, 0.75, 0.1], [0.75, 0.75, 0.1], "cpu"),
        (
            [-1, -1, -1],
            [1, 1, 1],
            [[-2, 1, 0.25], [1.5, -1, 0.75]],
            [[-1, 1, 0.25], [1, -1, 0.75]],
            "cpu",
        ),
        (0, 1, [1.1, 0.75, -1], [1.0, 0.75, 0], "cuda"),
        ([0.5, 0, 0.1], 1, [0, 0, 0.2], [0.5, 0, 0.2], "cuda"),
        (0, [0.75, 1.0, 0.1], [1.0, 0.75, 0.1], [0.75, 0.75, 0.1], "cuda"),
        (
            [-1, -1, -1],
            [1, 1, 1],
            [[-2, 1, 0.25], [1.5, -1, 0.75]],
            [[-1, 1, 0.25], [1, -1, 0.75]],
            "cuda",
        ),
    ],
)
def test_multi_dim_clamp(min, max, action, expected_result, vector_space, device):
    if device == "cuda" and not torch.cuda.is_available():
        pytest.skip("CUDA not available")

    if isinstance(min, list):
        min = np.array(min)
    if isinstance(max, list):
        max = np.array(max)

    ddpg = DDPG(
        observation_space=vector_space,
        action_space=spaces.Box(low=-1, high=1, shape=(1,), dtype=np.float32),
        device=device,
    )
    input = torch.tensor(action, dtype=torch.float32).to(device)
    clamped_actions = ddpg.multi_dim_clamp(min, max, input).type(torch.float32)
    expected_result = torch.tensor(expected_result)
    assert clamped_actions.dtype == expected_result.dtype
    ddpg.clean_up()



from datetime import datetime

import numpy as np
import torch
import wandb
from gymnasium import spaces
from tensordict import TensorDict
from ucimlrepo import fetch_ucirepo

from agilerl.algorithms import NeuralUCB
from agilerl.components.replay_buffer import ReplayBuffer
from agilerl.hpo.mutation import Mutations
from agilerl.hpo.tournament import TournamentSelection
from agilerl.utils.utils import create_population, default_progress_bar
from agilerl.wrappers.learning import BanditEnv

# !Note: If you are running this demo without having installed agilerl,
# uncomment and place the following above agilerl imports:

# import sys
# sys.path.append('../')


if __name__ == "__main__":
    print("===== AgileRL Bandit Demo =====")

    device = torch.device("cuda" if torch.cuda.is_available() else "cpu")

    NET_CONFIG = {
        "encoder_config": {
            "encoder_config": {"hidden_size": [128]}  # Actor hidden size
        }
    }

    INIT_HP = {
        "BATCH_SIZE": 64,  # Batch size
        "LR": 1e-3,  # Learning rate
        "GAMMA": 1.0,  # Scaling factor
        "LAMBDA": 1.0,  # Regularization factor
        "REG": 0.000625,  # Loss regularization factor
        "LEARN_STEP": 2,  # Learning frequency
        "POP_SIZE": 4,  # Population size
    }

    # Fetch data  https://archive.ics.uci.edu/
    iris = fetch_ucirepo(id=53)
    features = iris.data.features
    targets = iris.data.targets

    env = BanditEnv(features, targets)  # Create environment
    context_dim = env.context_dim
    action_dim = env.arms

    observation_space = spaces.Box(
        low=features.values.min(), high=features.values.max()
    )
    action_space = spaces.Discrete(env.arms)
    pop: list[NeuralUCB] = create_population(
        algo="NeuralUCB",  # Algorithm
        observation_space=observation_space,  # State dimension
        action_space=action_space,  # Action dimension
        net_config=NET_CONFIG,  # Network configuration
        INIT_HP=INIT_HP,  # Initial hyperparameters
        population_size=INIT_HP["POP_SIZE"],  # Population size
        device=device,
    )

    memory = ReplayBuffer(INIT_HP["MEMORY_SIZE"], device=device)

    tournament = TournamentSelection(
        tournament_size=2,  # Tournament selection size
        elitism=True,  # Elitism in tournament selection
        population_size=INIT_HP["POP_SIZE"],  # Population size
        eval_loop=1,  # Evaluate using last N fitness scores
    )
    mutations = Mutations(
        algo="NeuralUCB",  # Algorithm
        no_mutation=0.4,  # No mutation
        architecture=0.2,  # Architecture mutation
        new_layer_prob=0.5,  # New layer mutation
        parameters=0.2,  # Network parameters mutation
        activation=0.2,  # Activation layer mutation
        rl_hp=0.2,  # Learning HP mutation
        rl_hp_selection=["lr", "batch_size"],  # Learning HPs to choose from
        mutation_sd=0.1,  # Mutation strength  # Network architecture
        rand_seed=1,  # Random seed
        device=device,
    )

    max_steps = 10000  # Max steps per episode
    episode_steps = 500  # Steps in episode
    evo_steps = 500  # Evolution frequency
    eval_steps = 500  # Evaluation steps per episode
    eval_loop = 1  # Number of evaluation episodes

    print("Training...")

    wandb.init(
        # set the wandb project where this run will be logged
        project="AgileRL-Bandits",
        name="NeuralUCB-{}".format(datetime.now().strftime("%m%d%Y%H%M%S")),
        # track hyperparameters and run metadata
        config=INIT_HP,
    )

    total_steps = 0
    evo_count = 0

    # TRAINING LOOP
    print("Training...")
    pbar = default_progress_bar(max_steps)
    while np.less([agent.steps[-1] for agent in pop], max_steps).all():
        pop_episode_scores = []
        for agent_idx, agent in enumerate(pop):  # Loop through population
            score = 0
            losses = []
            context = env.reset()  # Reset environment at start of episode
            for idx_step in range(episode_steps):
                # Get next action from agent
                action = agent.get_action(context)
                next_context, reward = env.step(action)  # Act in environment

                # Save experience to replay buffer
                transition = TensorDict(
                    {
                        "obs": context[action],
                        "reward": reward,
                    },
                ).float()
                transition.batch_size = [1]
                memory.add(transition)

                # Learn according to learning frequency
                if len(memory) >= agent.batch_size:
                    for _ in range(agent.learn_step):
                        # Sample replay buffer
                        # Learn according to agent's RL algorithm
                        experiences = memory.sample(agent.batch_size)
                        loss = agent.learn(experiences)
                        losses.append(loss)

                context = next_context
                score += reward
                agent.regret.append(agent.regret[-1] + 1 - reward)

            agent.scores.append(score)
            pop_episode_scores.append(score)
            agent.steps[-1] += episode_steps
            total_steps += episode_steps
            pbar.update(episode_steps // len(pop))

            wandb_dict = {
                "global_step": total_steps,
                "train/loss": np.mean(losses),
                "train/score": score,
                "train/mean_regret": np.mean([agent.regret[-1] for agent in pop]),
            }
            wandb.log(wandb_dict)

        # Evaluate population
        fitnesses = [
            agent.test(
                env,
                max_steps=eval_steps,
                loop=eval_loop,
            )
            for agent in pop
        ]

        pbar.write(
            f"--- Global steps {total_steps} ---\n"
            f"Steps: {[agent.steps[-1] for agent in pop]}\n"
            f"Regret: {[agent.regret[-1] for agent in pop]}\n"
            f"Fitnesses: {['%.2f' % fitness for fitness in fitnesses]}\n"
            f"5 fitness avgs: {['%.2f' % np.mean(agent.fitness[-5:]) for agent in pop]}\n"
        )

        if pop[0].steps[-1] // evo_steps > evo_count:
            # Tournament selection and population mutation
            elite, pop = tournament.select(pop)
            pop = mutations.mutation(pop)
            evo_count += 1

        # Update step counter
        for agent in pop:
            agent.steps.append(agent.steps[-1])

    pbar.close()
    env.close()



from .base import EvolvableAlgorithm, LLMAlgorithm, MultiAgentRLAlgorithm, RLAlgorithm
from .optimizer_wrapper import OptimizerWrapper

__all__ = [
    "EvolvableAlgorithm",
    "RLAlgorithm",
    "MultiAgentRLAlgorithm",
    "OptimizerWrapper",
    "LLMAlgorithm",
]
